import unittest
from typing import List

from pyspark.sql import DataFrame
from pyspark.sql.types import IntegerType, StringType, StructField, StructType
from spetlrtools.testing import TestHandle

from spetlr.etl import Transformer
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.types import dataset_group
from spetlr.spark import Spark
from tests.local.etl.etl_transformer_test_helper import ETLTransformerTester


class TransformerTests(unittest.TestCase):
    """
    Test suite for the Transformer class in the OETL framework.

    This test suite contains test cases for both
        consuming and non-consuming transformers.
    """

    @classmethod
    def setUpClass(cls):
        cls.transformer_consuming = TestTransformer(consume_inputs=True)
        cls.transformer_non_consuming = TestTransformer(consume_inputs=False)
        cls.df = create_dataframe()

        # Test helpers for non consuming handling of keys
        cls.transformer_output_key = TestTransformer(
            dataset_output_key="my_output_key", consume_inputs=False
        )
        cls.transformer_input_key = TestTransformer(
            dataset_input_keys=["my_input_key"], consume_inputs=False
        )
        cls.transformer_input_key_list = TestTransformer(
            dataset_input_keys=["my_input_key_1", "my_input_key_2"],
            consume_inputs=False,
        )

    # ######################################################################
    # ################# CONSUMING TEST CASES ###############################
    # ######################################################################
    def test_consuming_process(self):
        # Process is called with single input
        # Result contain single df
        result = self.transformer_consuming.etl({"df": self.df})
        self.assertEqual(1, len(result))
        self.assertEqual({"TestTransformer"}, set(result.keys()))
        self.assertIs(self.df, result["TestTransformer"])

    def test_consuming_process_many(self):
        # Process_many is called with multiple inputs
        # Result contain union of input dfs
        result = self.transformer_consuming.etl({"df1": self.df, "df2": self.df})
        self.assertEqual(1, len(result))
        self.assertEqual({"TestTransformer"}, set(result.keys()))
        self.assertEqual(
            self.df.union(self.df).collect(), result["TestTransformer"].collect()
        )

    def test_consuming_dataset_key(self):
        # Set dataset_key to control name of the output
        # Result contain the key 'my_key'
        result = TestTransformer(dataset_key="my_key").etl({"df": self.df})
        self.assertEqual(1, len(result))
        self.assertEqual({"my_key"}, set(result.keys()))
        self.assertIs(self.df, result["my_key"])

    def test_consuming_dataset_output_key(self):
        # Set dataset_output_key to control name of the output
        # Result contain the key 'my_key'
        result = TestTransformer(dataset_output_key="my_key").etl({"df": self.df})
        self.assertEqual(1, len(result))
        self.assertEqual({"my_key"}, set(result.keys()))
        self.assertIs(self.df, result["my_key"])

    def test_consuming_single_dataset_input_key(self):
        # Set dataset_input_keys to control what dfs in dataset to handle
        # Process is called with single input
        # Result contain two keys due to 'df2' not being consumed
        # when not in dataset_input_keys
        result = TestTransformer(dataset_input_keys=["df1"]).etl(
            {"df1": self.df, "df2": self.df}
        )
        self.assertEqual(2, len(result))
        self.assertEqual({"TestTransformer", "df2"}, set(result.keys()))
        self.assertIs(self.df, result["TestTransformer"])
        self.assertIs(self.df, result["df2"])

    def test_consuming_multiple_dataset_input_keys(self):
        # Set dataset_input_keys to control what dfs in dataset to handle
        # Process_many is called with multiple inputs
        # Result contain two keys due to 'df3' not being consumed
        # when not in dataset_input_keys
        result = TestTransformer(dataset_input_keys=["df1", "df2"]).etl(
            {"df1": self.df, "df2": self.df, "df3": self.df}
        )
        self.assertEqual(2, len(result))
        self.assertEqual({"TestTransformer", "df3"}, set(result.keys()))
        self.assertEqual(
            self.df.union(self.df).collect(), result["TestTransformer"].collect()
        )
        self.assertIs(self.df, result["df3"])

    def test_consuming_should_not_consume_inputs(self):
        # Set consume_inputs to False to control dataset keys will no be consumed
        # in the output
        # Result contain two keys due to 'df' not being consumed
        result = TestTransformer(consume_inputs=False).etl({"df": self.df})
        self.assertEqual(2, len(result))
        self.assertEqual({"TestTransformer", "df"}, set(result.keys()))
        self.assertIs(self.df, result["TestTransformer"])
        self.assertIs(self.df, result["df"])

    def test_consuming_in_etl_context(self):
        """
        In order to test whether the transformer class
        works in a full ETL flow,
        this tests a simple setup.

        It should be able to write df_trans to a testhandle.

        But cannot write df_1 to a testhandle, since it has been consumed.
        """
        trans_consuming = TestTransformer(
            dataset_input_keys=["df_1"],
            dataset_output_key="df_trans",
            consume_inputs=True,
        )

        oc_test_B = ETLTransformerTester(
            trans_consuming,
        )
        dh_load = TestHandle()

        oc_test_B.load_into(
            SimpleLoader(
                handle=dh_load, mode="overwrite", dataset_input_keys=["df_trans"]
            )
        )

        oc_test_B.execute()

        oc_test_B.load_into(
            SimpleLoader(handle=dh_load, mode="overwrite", dataset_input_keys=["df_1"])
        )

        with self.assertRaises(KeyError) as cm:
            oc_test_B.execute()

        # Since it is consuming, it should not be able to find df_1
        self.assertEqual(cm.exception.args[0], "df_1")

    # ######################################################################
    # ################# NON-CONSUMING TEST CASES ###########################
    # ######################################################################

    def test_non_consuming_process(self):
        # process is called when:
        # - inputs has one df
        # - dataset_input_key is not given
        result = self.transformer_non_consuming.etl({"df": self.df})
        self.assertIs(self.df, list(result.values())[1])

    def test_non_consuming_process_many(self):
        # process_many is called when:
        # - inputs has more than one df
        # - dataset_input_key is not given
        result = self.transformer_non_consuming.etl({"df1": self.df, "df2": self.df})
        self.assertEqual(6, list(result.values())[2].count())

    def test_non_consuming_process_no_key(self):
        # process sets dataset_ouput_key equivalent to the transformer class when:
        # - dataset_output_key is not given
        result = self.transformer_non_consuming.etl({"df": self.df})
        self.assertEqual("TestTransformer", list(result.keys())[1])

    def test_non_consuming_process_many_no_key(self):
        # process_many sets dataset_ouput_key equivalent to the transformer class when:
        # - dataset_output_key is not given
        result = self.transformer_non_consuming.etl({"df1": self.df, "df2": self.df})
        self.assertEqual("TestTransformer", list(result.keys())[2])

    def test_non_consuming_process_output_key(self):
        # process sets dataset_output_key equiavalent to the input when:
        # - dataset_output_key is given
        result = self.transformer_output_key.etl({"df": self.df})
        self.assertEqual("my_output_key", list(result.keys())[1])

    def test_non_consuming_process_many_output_key(self):
        # process_many sets dataset_output_key equiavalent to the input when:
        # - dataset_output_key is given
        result = self.transformer_output_key.etl({"df1": self.df, "df2": self.df})
        self.assertEqual("my_output_key", list(result.keys())[2])

    def test_non_consuming_process_input_key(self):
        # process is called when:
        # - dataset_input_key is given
        result = self.transformer_input_key.etl({"my_input_key": self.df})
        self.assertIs(self.df, list(result.values())[1])

        with self.assertRaises(KeyError):
            # error when:
            # - the provided dataset_input_key does not exist in inputs
            self.transformer_input_key.etl({"df": self.df})

    def test_non_consuming_process_many_input_key_list(self):
        # process_many is called when:
        # - dataset_input_key_list is given
        result = self.transformer_input_key_list.etl(
            {"my_input_key_1": self.df, "my_input_key_2": self.df}
        )
        self.assertEqual(6, list(result.values())[2].count())

        with self.assertRaises(AssertionError):
            # error when:
            # - the provided dataset_input_key_list contains an incorrect key
            self.transformer_input_key_list.etl({"df1": self.df, "df2": self.df})

    def test_non_consuming_in_etl_context(self):
        """
        In order to test whether the transformer class
        works in a full ETL flow,
        this tests a simple setup.

        The non consuming can both write the df_1 and df_trans
        to a testhandle.
        """

        trans_consuming = TestTransformer(
            dataset_input_keys=["df_1"],
            dataset_output_key="df_trans",
            consume_inputs=False,
        )
        dh_load = TestHandle()
        oc_test_A = ETLTransformerTester(trans_consuming)
        oc_test_A.load_into(
            SimpleLoader(
                handle=dh_load, mode="overwrite", dataset_input_keys=["df_trans"]
            )
        )
        oc_test_A.load_into(
            SimpleLoader(handle=dh_load, mode="overwrite", dataset_input_keys=["df_1"])
        )

        oc_test_A.execute()


class TestTransformer(Transformer):
    def __init__(
        self,
        dataset_key: str = None,
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True,
    ):
        super().__init__(
            dataset_key=dataset_key,
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )

    def process(self, df: DataFrame) -> DataFrame:
        return df

    def process_many(self, datasets: dataset_group) -> DataFrame:
        assert len(datasets) == 2
        df1, df2 = list(datasets.values())
        return df1.union(df2)


def create_dataframe():
    data = [(1, "1"), (2, "2"), (3, "3")]

    schema = StructType(
        [
            StructField("id", IntegerType(), False),
            StructField("text", StringType(), False),
        ]
    )

    return Spark.get().createDataFrame(data=data, schema=schema)


if __name__ == "__main__":
    unittest.main()



import unittest
import uuid

from spetlr import Configurator
from spetlr.cosmos import CosmosDb
from spetlr.functions import init_dbutils
from spetlr.spark import Spark


class CosmosTests(unittest.TestCase):
    _cm: CosmosDb
    _tc: Configurator

    @classmethod
    def setUpClass(cls):
        dbutils = init_dbutils()

        cls._cm = CosmosDb(
            endpoint=dbutils.secrets.get("secrets", "Cosmos--Endpoint"),
            account_key=dbutils.secrets.get("secrets", "Cosmos--AccountKey"),
            database=f"SpetlrCosmosContainer{uuid.uuid4().hex}",
            catalog_name=None,
        )

        cls._tc = Configurator()
        cls._tc.clear_all_configurations()
        cls._tc.set_debug()
        cls._tc.register(
            "CmsTbl",
            {
                "name": "CosmosTable{ID}",
                "schema": {"sql": "id string, pk string, value int"},
                "rows_per_partition": 5,
            },
        )

    def test_01_create_db(self):
        self._cm.create_table(
            table_name=self._tc.table_name("CmsTbl"),
            partition_key="/pk",
            offer_throughput=400,
        )

    def test_02_write_table(self):
        df = Spark.get().createDataFrame(
            [("first", "pk1", 56), ("second", "pk2", 987)],
            "id string, pk string, value int",
        )
        self._cm.write_table(df, "CmsTbl")

    def test_03_read_table(self):
        df = self._cm.read_table("CmsTbl")
        data = set(tuple(row) for row in df.collect())
        self.assertEqual({("first", "pk1", 56), ("second", "pk2", 987)}, data)

    def test_04_use_handle(self):
        ch = self._cm.from_tc("CmsTbl")
        df = ch.read()
        data = set(tuple(row) for row in df.collect())
        self.assertEqual({("first", "pk1", 56), ("second", "pk2", 987)}, data)

        # create a df with two new rows.
        new_df = Spark.get().createDataFrame(
            [("third", "pk1", 56), ("fourth", "pk2", 987)],
            "id string, pk string, value int",
        )

        # in append mode, there will now be 4 rows.
        ch.append(new_df)
        self.assertEqual(ch.read().count(), 4)

        # drop and recreate the container
        # we should end up with only the new rows.
        ch.recreate()
        ch.append(new_df)
        self.assertEqual(ch.read().count(), 2)

    @classmethod
    def tearDownClass(cls):
        cls._cm.delete_item("CmsTbl", "third", "pk1")
        cls._cm.delete_item("CmsTbl", "fourth", "pk2")

        cls._cm.client.delete_database(cls._cm.database)



import unittest
from datetime import datetime

import pyspark.sql.types as T

from spetlr.spark import Spark
from spetlr.transformations import join_time_series_dataframes


class TestFunctions(unittest.TestCase):
    def test_join_time_series_dataframes(self):
        spark = Spark.master("local[*]").get()

        df1Schema = T.StructType(
            [
                T.StructField("Id", T.LongType(), True),
                T.StructField("Name", T.StringType(), True),
                T.StructField("StartTimestamp", T.TimestampType(), True),
                T.StructField("EndTimestamp", T.TimestampType(), True),
                T.StructField("State", T.StringType(), True),
            ]
        )

        df1Data = [
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=1, hour=0),
                datetime(year=2020, month=1, day=2, hour=0),
                "state1",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=2, hour=0),
                datetime(year=2020, month=1, day=3, hour=0),
                "state2",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=3, hour=0),
                datetime(year=2020, month=1, day=4, hour=0),
                "state1",
            ),
        ]

        df2Schema = T.StructType(
            [
                T.StructField("Id", T.LongType(), True),
                T.StructField("StartTimestamp", T.TimestampType(), True),
                T.StructField("EndTimestamp", T.TimestampType(), True),
                T.StructField("State", T.StringType(), True),
            ]
        )

        df2Data = [
            (
                1,
                datetime(year=2020, month=1, day=1, hour=6),
                datetime(year=2020, month=1, day=1, hour=8),
                "state3",
            ),
            (
                1,
                datetime(year=2020, month=1, day=1, hour=10),
                datetime(year=2020, month=1, day=1, hour=12),
                "state3",
            ),
            (
                1,
                datetime(year=2020, month=1, day=1, hour=20),
                datetime(year=2020, month=1, day=2, hour=4),
                "state4",
            ),
            (
                1,
                datetime(year=2020, month=1, day=2, hour=6),
                datetime(year=2020, month=1, day=2, hour=8),
                "state3",
            ),
            (
                1,
                datetime(year=2020, month=1, day=2, hour=10),
                datetime(year=2020, month=1, day=2, hour=12),
                "state3",
            ),
        ]

        dfExpectedData = [
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=1, hour=0),
                datetime(year=2020, month=1, day=1, hour=6),
                "state1",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=1, hour=6),
                datetime(year=2020, month=1, day=1, hour=8),
                "state3",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=1, hour=8),
                datetime(year=2020, month=1, day=1, hour=10),
                "state1",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=1, hour=10),
                datetime(year=2020, month=1, day=1, hour=12),
                "state3",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=1, hour=12),
                datetime(year=2020, month=1, day=1, hour=20),
                "state1",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=1, hour=20),
                datetime(year=2020, month=1, day=2, hour=4),
                "state4",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=2, hour=4),
                datetime(year=2020, month=1, day=2, hour=6),
                "state2",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=2, hour=6),
                datetime(year=2020, month=1, day=2, hour=8),
                "state3",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=2, hour=8),
                datetime(year=2020, month=1, day=2, hour=10),
                "state2",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=2, hour=10),
                datetime(year=2020, month=1, day=2, hour=12),
                "state3",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=2, hour=12),
                datetime(year=2020, month=1, day=3, hour=0),
                "state2",
            ),
            (
                1,
                "TestName",
                datetime(year=2020, month=1, day=3, hour=0),
                datetime(year=2020, month=1, day=4, hour=0),
                "state1",
            ),
        ]

        # Construct dataframes
        df1 = spark.createDataFrame(df1Data, df1Schema)
        df2 = spark.createDataFrame(df2Data, df2Schema)
        dfExpected = spark.createDataFrame(dfExpectedData, df1Schema)

        dfReturn = join_time_series_dataframes(
            dfPrimary=df1,
            dfSecondary=df2,
            startTimeColumnName="StartTimestamp",
            endTimeColumnName="EndTimestamp",
            idColumns=["Id"],
            stateColumn="State",
        )

        self.assertEqual(
            dfExpected.orderBy("StartTimestamp").collect(),
            dfReturn.orderBy("StartTimestamp").collect(),
        )


if __name__ == "__main__":
    unittest.main()



from spetlr.delta import DeltaHandle
from spetlr.eh.EventHubCaptureExtractor import EventHubCaptureExtractor
from spetlr.etl import EtlBase, Orchestrator
from spetlr.etl.loaders import SimpleLoader
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaExtractor import (
    EhJsonToDeltaExtractor,
)
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaTransformer import (
    EhJsonToDeltaTransformer,
)


class EhJsonToDeltaOrchestrator(Orchestrator):
    def __init__(
        self,
        eh: EventHubCaptureExtractor,
        dh: DeltaHandle,
        *,
        case_sensitive: bool = True,
    ):
        super().__init__()
        self.eh = eh
        self.dh = dh

        # step 1,
        #  - get the highest partition from the delta table,
        #  - truncate that partition
        #  - read the capture files from that partition
        self.extract_from(EhJsonToDeltaExtractor(eh, dh))

        # step 2,
        #  - use the target schema to select what to copy from capture files
        #  - anything that is not in the source df is used to unpack the body json
        self.transform_with(
            EhJsonToDeltaTransformer(target_dh=dh, case_sensitive=case_sensitive)
        )

        # the method filter_with can be used to insert any number of transformers here

        # final step,
        # append the rows to the delta table.
        self._loader = SimpleLoader(dh, mode="append")
        self.load_into(self._loader)

    @classmethod
    def from_tc(cls, eh_id: str, dh_id: str):
        return cls(
            eh=EventHubCaptureExtractor.from_tc(eh_id), dh=DeltaHandle.from_tc(dh_id)
        )

    def filter_with(self, etl: EtlBase):
        """Additional filters to execute before loading."""
        loader = self.steps.pop()

        # the following will fail if additional steps have been added.
        assert self._loader is loader, "unexpected change in etl steps"

        self.transform_with(etl)
        self.load_into(loader)



import pyspark.sql.types
from deprecated import deprecated
from pyspark.sql import DataFrame
from pyspark.sql.utils import AnalysisException

from spetlr import Configurator
from spetlr.configurator.sql.parse_sql import parse_single_sql_statement
from spetlr.delta import DeltaHandle
from spetlr.deltaspec.DeltaTableSpecBase import (
    DeltaTableSpecBase,
    _DEFAULT_blankedPropertyKeys,
)
from spetlr.deltaspec.DeltaTableSpecDifference import DeltaTableSpecDifference
from spetlr.deltaspec.exceptions import (
    InvalidSpecificationError,
    NoTableAtTarget,
    TableSpecNotReadable,
)
from spetlr.schema_manager.spark_schema import get_schema
from spetlr.spark import Spark
from spetlr.sqlrepr.sql_types import repr_sql_types


@deprecated("Class untested in current version of spetlr")
class DeltaTableSpec(DeltaTableSpecBase):
    """This class represents a full specification for a delta table."""

    # Non-trivial constructors
    @classmethod
    def from_sql(cls, sql: str) -> "DeltaTableSpec":
        """Return the DeltaTableSpec instance,
        that describes the table that would be created
        by the sql CREATE TABLE statement in the argument."""
        details = parse_single_sql_statement(sql)
        if details.get("format").lower() != "delta":
            raise InvalidSpecificationError(
                "The sql code is not a create table statement."
            )
        schema = get_schema(details["schema"]["sql"])
        init_args = dict(
            name=details.get("name"),
            location=details.get("path"),
            comment=details.get("comment"),
            schema=schema,
            options=details.get("options", {}),
            partitioned_by=details.get("partitioned_by", []),
            cluster_by=details.get("cluster_by", []),
            tblproperties=details.get("tblproperties", {}),
        )

        init_args = {k: v for k, v in init_args.items() if v}

        o = cls(**init_args)

        # objects initialized from parsed SQL should have these set
        # so that we define capable tables.
        o.set_good_defaults()

        return o

    @classmethod
    def from_path(cls, location: str) -> "DeltaTableSpec":
        """Return the DeltaTableSpec instance,
        that describes the table found at the given storage location.
        The name will be None since that is not stored in the data."""
        return cls.from_name(f"delta.`{location}`")

    @classmethod
    def from_tc(cls, id: str) -> "DeltaTableSpec":
        c = Configurator()
        # schema is required
        item = c._get_item(id)
        schema = get_schema(item["schema"]["sql"])
        init_args = dict(
            name=item.get("name", None),
            location=item.get("path", None),
            comment=item.get("comment", None),
            schema=schema,
            options=item.get("options", {}),
            partitioned_by=item.get("partitioned_by", []),
            cluster_by=item.get("cluster_by", []),
            tblproperties=item.get("tblproperties", {}),
        )

        init_args = {k: v for k, v in init_args.items() if v}

        o = cls(**init_args)

        # objects initialized from parsed SQL should have these set
        # so that we define capable tables.
        o.set_good_defaults()

        return o

    @classmethod
    def from_name(cls, in_name: str) -> "DeltaTableSpec":
        """Return the DeltaTableSpec instance,
        that describes the table of the given name."""
        spark = Spark.get()
        try:
            details = spark.sql(f"DESCRIBE DETAIL {in_name}").collect()[0].asDict()

        except AnalysisException as e:
            raise NoTableAtTarget(str(e))

        if details["format"] != "delta":
            raise InvalidSpecificationError("The table is not of delta format.")

        tblproperties = details["properties"]
        tblproperties["delta.minReaderVersion"] = str(details["minReaderVersion"])
        tblproperties["delta.minWriterVersion"] = str(details["minWriterVersion"])

        return cls(
            name=details["name"],
            schema=spark.table(in_name).schema,
            partitioned_by=details["partitionColumns"],
            cluster_by=details["clusteringColumns"],
            tblproperties=tblproperties,
            location=details["location"],
            comment=details["description"],
        )

    # String representations

    def __repr__(self):
        """Return a correct and minimal string,
         that can be evaluated as python code to return a DeltaTableSpec instance
        that will compare equal to the current instance."""
        parts = [
            (f"name={repr(self.name)}"),
            f"schema={repr_sql_types(self.schema)}",
            (f"options={repr(self.options)}" if self.options else ""),
            (
                f"partitioned_by={repr(self.partitioned_by)}"
                if self.partitioned_by
                else ""
            ),
            (f"cluster_by={repr(self.cluster_by)}" if self.cluster_by else ""),
            (f"tblproperties={repr(self.tblproperties)}" if self.tblproperties else ""),
            (f"comment={repr(self.comment)}" if self.comment else ""),
            (f"location={repr(self.location)}" if self.location else ""),
            (
                f"blankedPropertyKeys={repr(self.blankedPropertyKeys)}"
                if self.blankedPropertyKeys != _DEFAULT_blankedPropertyKeys
                else ""
            ),
        ]

        return "DeltaTableSpec(" + (", ".join(p for p in parts if p)) + ")"

    # identity manipulation
    def copy(self) -> "DeltaTableSpec":
        """Return an independent object that compares equal to this one."""
        globals = {
            k: v for k, v in vars(pyspark.sql.types).items() if not k.startswith("_")
        }
        globals.update(dict(DeltaTableSpec=DeltaTableSpec))
        return eval(repr(self), globals)

    def get_dh(self) -> DeltaHandle:
        full = self.fully_substituted()
        return DeltaHandle(name=full.name, location=full.location, data_format="delta")

    def compare_to(self, other: "DeltaTableSpec") -> DeltaTableSpecDifference:
        """Returns a DeltaTableSpecDifference
        of what this difference this object has with respect to the other."""
        return DeltaTableSpecDifference(base=other, target=self)

    def compare_to_location(self):
        """Returns a DeltaTableSpecDifference of self with respect to the disk."""
        return DeltaTableSpec.compare_to_name(self.fully_substituted(name=None))

    def compare_to_name(self):
        """Returns a DeltaTableSpecDifference of self
        with respect to the catalog table of the same name."""
        full = self.fully_substituted()
        try:
            if full.name:
                onstorage = DeltaTableSpec.from_name(full.name)
            else:
                onstorage = DeltaTableSpec.from_path(full.location)
        except NoTableAtTarget:
            onstorage = None
        return DeltaTableSpecDifference(base=onstorage, target=full)

    def make_storage_match(
        self,
        allow_columns_add=False,
        allow_columns_drop=False,
        allow_columns_type_change=False,
        allow_columns_reorder=False,
        allow_name_change=False,
        allow_location_change=False,
        allow_table_create=True,
        errors_as_warnings=False,
    ) -> None:
        """If storage is not exactly like the specification,
        change the storage to make it match."""
        diff = self.compare_to_name()

        if diff.is_different():
            spark = Spark.get()
            print(f"Now altering table {diff.target.name} to match specification:")
            for statement in diff.alter_statements(
                allow_columns_add=allow_columns_add,
                allow_columns_drop=allow_columns_drop,
                allow_columns_type_change=allow_columns_type_change,
                allow_columns_reorder=allow_columns_reorder,
                allow_name_change=allow_name_change,
                allow_location_change=allow_location_change,
                allow_table_create=allow_table_create,
                errors_as_warnings=errors_as_warnings,
            ):
                print(f"Executing SQL: {statement}")
                spark.sql(statement)

    def ensure_df_schema(self, df: DataFrame):
        # check if the df can be selected down into the schema of this table
        if not self.compare_to(
            DeltaTableSpec(schema=df.schema, name=None, location=self.location)
        ).is_readable():
            raise TableSpecNotReadable(
                "The data frame has an incompatible schema mismatch to this table."
            )
        return df.select(*self.schema.names)

    def is_readable(self) -> bool:
        """Check if this delta table is readable when compared
        to the table of the same name currently in spark.
        Readability means that all columns in the schema of
        this specification also exist in table in spark.
        """
        return self.compare_to_name().is_readable()



from spetlr.config_master import Configurator
from spetlr.exceptions import OnlyUseInSpetlrDebugMode
from spetlr.spark import Spark


def CleanupTestDatabases(catalog: str = None):
    """
    This function can be applied for removing test databases
    in the Databricks Environment.
    """

    c = Configurator()
    if not c.is_debug():
        raise OnlyUseInSpetlrDebugMode()

    ID = c.get("ID")

    if catalog is None:
        sql = f"SHOW DATABASES LIKE '*{ID}'"
    else:
        sql = f"SHOW DATABASES IN {catalog} LIKE '*{ID}'"

    for (db,) in Spark.get().sql(sql).collect():
        if catalog:
            db = f"{catalog}.{db}"
        print(f"Now deleting database {db}")
        Spark.get().sql(f"DROP DATABASE {db} CASCADE")
    print("Database cleanup done.")



import pyspark.sql.functions as F
from pyspark.sql import DataFrame

from spetlr.etl.log import LogTransformer


class CountLogTransformer(LogTransformer):
    """
    A subclass of LogTransformer that logs the number of rows in a DataFrame.

    Arguments:
        log_name (str): The name for the log entry.
        dataset_input_keys (List[str], optional): A list of input dataset keys
            that this transformer will process. Defaults to None.
        dataset_output_key (str, optional): The key for the output dataset
            generated by this transformer. If not provided, a random UUID will
            be used as the default value. It is not required for when using the
            LogOrchestrator.
        consume_inputs (bool, optional): Flag to control dataset consuming behavior.
            Defaults to True
    """

    def log(self, df: DataFrame) -> DataFrame:
        return df.select(F.count("*").alias("Count"))



import sys
from typing import List, Optional

import pyspark.sql.functions as f
from pyspark.sql import DataFrame

from spetlr.etl import Loader
from spetlr.spark import Spark

from ..functions import get_unique_tempview_name
from .CachedLoaderParameters import CachedLoaderParameters


class CachedLoader(Loader):
    """
    This class is an ETL loader that uses a cache table to turn
    a full load of a dataframe into an incremental load, where only new
    or changed rows are passed on. Deleted rows are also passed on but
    to a different method.
    The cache table needs to exist and have the correct schema matching
    the configured parameters, see the CachedLoaderParameters class.

    Remember to override self.write_operation and/or self.delete_operation.
    If the parameter "do_nothing_if_more_rows_than" was specified,
       you also need to override self.too_many_rows.
    Any variable rows need to be added in this function.
    """

    class ReductionResult:
        to_be_written: DataFrame
        to_be_deleted: DataFrame

    def __init__(self, params: CachedLoaderParameters):
        super().__init__()
        self.params = params
        self.validate()

    def validate(self):
        # validate cache table schema
        p = self.params
        df = Spark.get().table(p.cache_table_name)
        try:
            assert set(df.columns) == set(
                p.key_cols + p.cache_id_cols + [p.rowHash, p.loadedTime, p.deletedTime]
            )
        except AssertionError:
            print(
                "ERROR: The cache table needs to have precisely the correct schema."
                " Found:",
                df.columns,
                file=sys.stderr,
            )
            raise

        # validate overloading
        write_not_overloaded = "write_operation" not in self.__class__.__dict__
        delete_not_overloaded = "delete_operation" not in self.__class__.__dict__
        too_many_rows_not_overloaded = "too_many_rows" not in self.__class__.__dict__
        if write_not_overloaded or delete_not_overloaded:
            raise AssertionError(
                "write_operation and delete_operation methods required"
            )
        if too_many_rows_not_overloaded and p.do_nothing_if_more_rows_than is not None:
            raise AssertionError("too_many_rows method required")

        if self.__class__ is CachedLoader:
            raise AssertionError("You should inherit from this class")

    def write_operation(self, df: DataFrame) -> Optional[DataFrame]:
        """Abstract Method to be overridden in child."""
        raise NotImplementedError()

    def delete_operation(self, df: DataFrame) -> Optional[DataFrame]:
        """Abstract Method to be overridden in child."""
        raise NotImplementedError()

    def too_many_rows(self) -> None:
        """Abstract Method to be overridden in child."""
        raise NotImplementedError()

    def save(self, df: DataFrame) -> None:
        in_cols = df.columns

        if not set(self.params.key_cols).issubset(in_cols):
            raise AssertionError(
                "The key columns must be given in the input dataframe."
            )

        # ensure that no duplicates exist. The keys have to be a unique set.
        df = df.dropDuplicates(self.params.key_cols)

        cache = self._extract_cache()

        result = self._discard_non_new_rows_against_cache(df, cache)

        if self.params.do_nothing_if_more_rows_than is not None:
            if result.to_be_written.count() > self.params.do_nothing_if_more_rows_than:
                self.too_many_rows()
                return

        # write branch
        df_written = self.write_operation(result.to_be_written)
        if df_written:
            # this line only executes if the line above does not raise an exception.
            df_written_cache_update = self._prepare_written_cache_update(
                df_written, in_cols
            )
            self._load_cache(df_written_cache_update)

        # delete branch
        df_deleted = self.delete_operation(result.to_be_deleted)
        if df_deleted:
            df_deleted_cache_update = self._prepare_deleted_cache_update(df_deleted)

            self._load_cache(df_deleted_cache_update)
            # this method is called a again separately in order to ensure that the
            # written cache is saved even if the delete operation fails.

        return

    def _extract_cache(self) -> DataFrame:
        # here we fix the version,
        # so we don't overwrite the cache before we might want to use it.
        version = (
            Spark.get()
            .sql(f"DESCRIBE HISTORY {self.params.cache_table_name} LIMIT 1")
            .select("version")
            .take(1)[0][0]
        )
        cache = Spark.get().sql(
            f"SELECT * FROM {self.params.cache_table_name} VERSION AS OF {version}"
            f" WHERE {self.params.deletedTime} IS NULL"
            f" AND {self.params.rowHash} IS NOT NULL"
            f" AND {self.params.loadedTime} IS NOT NULL"
        )

        return cache

    def _prepare_written_cache_update(self, df: DataFrame, columns_to_hash: List[str]):
        # re-hash the input row
        return df.select(
            *self.params.key_cols,
            f.hash(*columns_to_hash).alias(self.params.rowHash),
            f.current_timestamp().alias(self.params.loadedTime),
            f.lit(None).cast("timestamp").alias(self.params.deletedTime),
            *self.params.cache_id_cols,
        )

    def _prepare_deleted_cache_update(self, df: DataFrame):
        return df.select(
            *self.params.key_cols,
            self.params.rowHash,
            self.params.loadedTime,
            f.current_timestamp().alias(self.params.deletedTime),
            *self.params.cache_id_cols,
        )

    def _load_cache(self, cache_to_load: DataFrame) -> None:
        view_name = get_unique_tempview_name()

        merge_sql_statement = (
            f"MERGE INTO {self.params.cache_table_name} AS target "
            f"USING {view_name} as source "
            f"ON "
            + (
                " AND ".join(
                    f"(source.{col} = target.{col})" for col in self.params.key_cols
                )
            )
            +
            # update existing records
            "WHEN MATCHED THEN UPDATE SET * "
            # insert new records.
            "WHEN NOT MATCHED THEN INSERT * "
        )
        cache_to_load.createOrReplaceTempView(view_name)
        Spark.get().sql(merge_sql_statement)

    def _discard_non_new_rows_against_cache(
        self, df_in: DataFrame, cache: DataFrame
    ) -> ReductionResult:
        """
        Returns:
            to_be_written contains only rows that
             - are new in the cache, or
             - are a mismatch against the cache
            to_be_deleted contains rows that
             - are present in the cache but not in the data
        """
        in_cols = df_in.columns

        # ensure no null keys:
        df_in = df_in.filter(
            " AND ".join(f"({col} is NOT NULL)" for col in self.params.key_cols)
        )

        # prepare hash of row
        df_hashed = df_in.withColumn("rowHash", f.hash("*"))

        # add a column to distinguish rows after the join
        df_hashed = df_hashed.withColumn("fromPayload", f.lit(True))

        # join with cache
        joined_df = (
            df_hashed.alias("df")
            .join(cache.alias("cache"), self.params.key_cols, "full")
            .cache()
        )

        result = self.ReductionResult()

        # rows coming from the original df will have a non-null fromPayload column
        result.to_be_written = (
            joined_df.filter("fromPayload IS NOT NULL")
            .filter(
                # either the row has never been loaded before
                (f.col(f"cache.{self.params.loadedTime}").isNull())
                # or it has changed wrt the previous load
                | (
                    f.col(f"df.{self.params.rowHash}")
                    != f.col(f"cache.{self.params.rowHash}")
                )
            )
            .select(
                *self.params.key_cols,
                *[f"df.{c}" for c in in_cols if c not in self.params.key_cols],
            )
        )

        # cached rows with no incoming match will have a null fromPayload column
        if Spark.version() > Spark.DATABRICKS_RUNTIME_9_1:
            # from databricks 10.4 this returns all columns, including key columns.
            # specifying key columns again causes them to be present twice in the
            # data frame
            result.to_be_deleted = joined_df.filter("fromPayload IS NULL").select(
                "cache.*"
            )
        else:
            # up to databricks 9.1 the key columns need to be selected separately,
            # since they are not returned by the cache.*
            result.to_be_deleted = joined_df.filter("fromPayload IS NULL").select(
                *self.params.key_cols, "cache.*"
            )

        return result



import unittest

import spetlr


class TestSparkImport(unittest.TestCase):
    def test_spark(self):
        spark = spetlr.spark.Spark.master("local[*]").get()
        df = spark.sql("select 42")
        self.assertEqual(42, df.collect()[0][0])


if __name__ == "__main__":
    unittest.main()



import time
from datetime import datetime, timedelta
from typing import Callable, Dict, List, Optional, Tuple

import msal
import pandas as pd
import requests
from pyspark.sql import DataFrame
from pytz import utc

from spetlr.power_bi.PowerBiClient import PowerBiClient
from spetlr.power_bi.PowerBiException import PowerBiException
from spetlr.power_bi.SparkPandasDataFrame import SparkPandasDataFrame


class PowerBi:
    def __init__(
        self,
        client: PowerBiClient,
        *,
        workspace_id: str = None,
        workspace_name: str = None,
        dataset_id: str = None,
        dataset_name: str = None,
        table_names: List[str] = None,
        max_minutes_after_last_refresh: int = 12 * 60,
        timeout_in_seconds: int = 15 * 60,
        number_of_retries: int = 0,
        max_parallelism: int = None,
        apply_refresh_policy: bool = None,
        dataset_refresh_type: str = None,
        dataset_commit_mode: str = None,
        incremental_effective_date: str = None,
        mail_on_failure: bool = False,
        mail_on_completion: bool = False,
        exclude_creators: List[str] = None,
        local_timezone_name: str = None,
        ignore_errors: bool = False,
    ):
        """
        Allows refreshing PowerBI datasets and checking if the last refresh
            completed successfully.
        If no workspace is specified, a list of available workspaces will be displayed.
        If no dataset is specified, a list of available datasets in the given workspace
            will be displayed.
        if the table names are an empty empty list, a list of available tables will
            be displayed.

        :param PowerBiClient client: PowerBI client credentials.

        :param str workspace_id: The GUID of the workspace.
        :param str workspace_name: The name of the workspace
            (specified instead of the workspace_id).
        :param str dataset_id: The GUID of the dataset.
        :param str dataset_name: The name of the dataset
            (specified instead of the dataset_id).
        :param list[str] table_names: Optional list of table names to be
            refreshed. By default all tables are refreshed and checked.
        :param int max_minutes_after_last_refresh: The number of minutes
            for which the last succeeded refresh is considered valid,
            or 0 to disable time checking. Default is 12 hours.
        :param bool timeout_in_seconds: The number of seconds after which
            the refresh() method times out. Default is 15 minutes.
        :param int number_of_retries: The number of retries on transient
            errors when calling refresh() and start_refresh().
            Default is 0 (no retries). (E.g. 1 means two attempts in total.)
            Used only when the timeout_in_seconds parameter allows it!
        :param int max_parallelism: Specifies the maximum number of threads
            that can run processing commands in parallel during a refresh
            in PowerBI. Default is None, which corresponds to 10 threads.
            Used only with the refresh() and start_refresh() methods!
        :param bool apply_refresh_policy: Determine if the refresh policy is applied
            or not (True or False). Default is None.
        :param str dataset_refresh_type: The type of processing to perform
            when refreshing the dataset
            ("Automatic", "Calculate", "ClearValues", "DataOnly", "Defragment", "Full").
            Default is None.
        :param str dataset_commit_mode: Determines if dataset objects will be
            committed in batches or only when complete
            ("PartialBatch", "Transactional"). Default is None.
        :param str incremental_effective_date: Effective date for incremental load.
            If an incremental refresh policy is applied, the effectiveDate parameter
            overrides the current date. Default is None.
        :param str local_timezone_name: The timezone to use when parsing
            timestamp columns. The default timezone is UTC.
            If the timezone is UTC, all timestamp columns will have a suffix "Utc".
            Otherwise they will have a suffix "Local".
        :param bool mail_on_failure: True to send an e-mail to the dataset
            owner when the refresh fails. Does not work with service principals!
        :param bool mail_on_completion: True to send an e-mail to the dataset
            owner when the refresh completes. Does not work with service principals!
        :param list[str] exclude_creators: When getting refresh histories/tables
            from several datasets simultaneously, exclude datasets configured
            by the specified creators (list of e-mails).
            This is to prevent the "Skipping unauthorized" message.
        :param bool ignore_errors: True to print errors in the output
            or False (default) to cast a PowerBiException.
        """

        if workspace_id is not None and workspace_name is not None:
            raise ValueError("Specify either 'workspace_id' or 'workspace_name'!")
        if dataset_id is not None and dataset_name is not None:
            raise ValueError("Specify either 'dataset_id' or 'dataset_name'!")
        if max_minutes_after_last_refresh is None or max_minutes_after_last_refresh < 0:
            raise ValueError(
                "The 'max_minutes_after_last_refresh' parameter "
                "must be greater than or equal zero!"
            )
        if timeout_in_seconds is None or timeout_in_seconds <= 0:
            raise ValueError(
                "The 'timeout_in_seconds' parameter must be greater than zero!"
            )
        if number_of_retries is None or number_of_retries < 0:
            raise ValueError(
                "The 'number_of_retries' parameter "
                "must be greater than or equal zero!"
            )
        if max_parallelism is not None and max_parallelism < 0:
            raise ValueError(
                "The 'max_parallelism' parameter must be greater than or equal zero!"
            )

        if (mail_on_failure or mail_on_completion) and table_names is not None:
            raise ValueError(
                "The options for refreshing selected tables "
                "and for sending notification emails cannot be combined!"
            )

        self.client = client
        self.workspace_id = workspace_id
        self.workspace_name = workspace_name
        self.dataset_id = dataset_id
        self.dataset_name = dataset_name
        self.table_names = table_names
        self.max_minutes_after_last_refresh = max_minutes_after_last_refresh
        self.timeout_in_seconds = timeout_in_seconds
        self.number_of_retries = number_of_retries
        self.max_parallelism = max_parallelism
        self.mail_on_failure = mail_on_failure
        self.mail_on_completion = mail_on_completion
        self.apply_refresh_policy = apply_refresh_policy
        self.dataset_refresh_type = dataset_refresh_type
        self.dataset_commit_mode = dataset_commit_mode
        self.incremental_effective_date = incremental_effective_date
        self.exclude_creators = (
            [(creator.lower() if creator else None) for creator in exclude_creators]
            if exclude_creators
            else []
        )
        self.local_timezone_name = local_timezone_name
        self.ignore_errors = ignore_errors

        self.api_header = None
        self.expire_time = 0
        self.last_status = None
        self.last_exception = None
        self.last_refresh_utc = None
        self.last_duration_in_seconds = 0

    def _raise_error(self, message: str) -> None:
        if self.ignore_errors:
            print(message)
        else:
            raise PowerBiException(message)

    def _raise_base_api_error(
        self,
        message: str,
        api_call: requests.Response,
        *,
        post_body: str = None,
        additional_message: str = None,
    ) -> None:
        print(api_call.text)
        print(api_call.url)
        if post_body is not None:
            print(post_body)
        if additional_message is not None:
            print(additional_message)
        self._raise_error(
            message + f" Response: {api_call.status_code} {api_call.reason}"
        )

    def _raise_api_error(
        self,
        message: str,
        api_call: requests.Response,
        *,
        post_body: str = None,
        workspace_id: str = None,
        dataset_id: str = None,
        request_id: str = None,
        ignore_unauthorized: bool = False,
    ) -> None:
        additional_message = ""
        if workspace_id is not None or dataset_id is not None or request_id is not None:
            additional_message = (
                f'workspace_id="{workspace_id}", dataset_id="{dataset_id}"'
            )
            if request_id is not None:
                additional_message += f', request_id="{request_id}"'
        if api_call.status_code == 404 and request_id is None:
            message = (
                "The specified dataset or workspace cannot be found, "
                "or the dataset doesn't have a user with the required permissions!"
            )
        if api_call.status_code == 401 and ignore_unauthorized:
            print("Skipping unauthorized: " + additional_message)
        else:
            self._raise_base_api_error(
                message,
                api_call,
                post_body=post_body,
                additional_message=additional_message,
            )

    def _get_access_token(self) -> bool:
        """
        Acquires an access token to connect to PowerBI.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if self.expire_time != 0:
            if time.time() < self.expire_time:
                return True
            print("Renewing access token...")

        # Prepare URLs
        authority_url = f"https://login.microsoftonline.com/{self.client.tenant_id}/"
        scope = ["https://analysis.windows.net/powerbi/api/.default"]
        self.powerbi_url = "https://api.powerbi.com/v1.0/myorg/"

        # Use MSAL to get an access token
        app = msal.ConfidentialClientApplication(
            self.client.client_id,
            authority=authority_url,
            client_credential=self.client.client_secret,
        )
        result = app.acquire_token_for_client(scopes=scope)
        if "access_token" not in result:
            print(result)
            self._raise_error("Failed to acquire token!")
            return False

        default_expires_in = 5 * 60
        extra_seconds = 3

        access_token = result["access_token"]
        if "expires_in" in result:
            self.expire_time = time.time() + result["expires_in"] - extra_seconds
        else:
            self.expire_time = time.time() + default_expires_in - extra_seconds

        self.api_header = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
        }
        return True

    def _get_workspaces(self) -> Optional[SparkPandasDataFrame]:
        """
        Returns the list of available PowerBI workspaces.

        :return: data frame with workspaces if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: SparkPandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        api_call = requests.get(
            url=f"{self.powerbi_url}groups", headers=self.api_header
        )
        if api_call.status_code == 200:
            return SparkPandasDataFrame(
                api_call.json()["value"],
                [
                    ("name", "WorkspaceName", "string"),
                    ("isReadOnly", "IsReadOnly", "boolean"),
                    ("isOnDedicatedCapacity", "IsOnDedicatedCapacity", "boolean"),
                    ("capacityId", "CapacityId", "string"),
                    (
                        "defaultDatasetStorageFormat",
                        "DefaultDatasetStorageFormat",
                        "string",
                    ),
                    ("dataflowStorageId", "DataflowStorageId", "string"),
                    ("id", "WorkspaceId", "string"),
                ],
                indexing_columns="id",
                sorting_columns="name",
            )
        self._raise_base_api_error("Failed to fetch workspaces!", api_call)
        return None

    def _get_datasets(self, workspace_id: str) -> Optional[SparkPandasDataFrame]:
        """
        Returns the list of available PowerBI datasets.

        :param str workspace_id: the id of the workspace to fetch
        :return: data frame with datasets if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: SparkPandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        api_call = requests.get(
            url=f"{self.powerbi_url}groups/{workspace_id}/datasets",
            headers=self.api_header,
        )
        if api_call.status_code == 200:
            return SparkPandasDataFrame(
                api_call.json()["value"],
                [
                    ("name", "DatasetName", "string"),
                    ("configuredBy", "ConfiguredBy", "string"),
                    ("isRefreshable", "IsRefreshable", "boolean"),
                    ("addRowsAPIEnabled", "AddRowsApiEnabled", "boolean"),
                    (
                        "isEffectiveIdentityRequired",
                        "IsEffectiveIdentityRequired",
                        "boolean",
                    ),
                    (
                        "isEffectiveIdentityRolesRequired",
                        "IsEffectiveIdentityRolesRequired",
                        "boolean",
                    ),
                    ("isOnPremGatewayRequired", "IsOnPremGatewayRequired", "boolean"),
                    ("id", "DatasetId", "string"),
                ],
                indexing_columns="id",
                sorting_columns=["name", "configuredBy"],
            )

        self._raise_base_api_error("Failed to fetch datasets!", api_call)
        return None

    def _get_refresh_history(
        self,
        workspace_id: str,
        dataset_id: str,
        ignore_unauthorized: bool = False,
    ) -> Optional[SparkPandasDataFrame]:
        """
        Returns the PowerBI dataset refresh history for the specified dataset.

        :param str workspace_id: the id of the workspace
        :param str dataset_id: the id of the dataset
        :param bool ignore_unauthorized: return None on HTTP 401 Unauthorized
        :return: data frame with the refresh history if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: SparkPandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        api_url = (
            f"{self.powerbi_url}groups/{workspace_id}/datasets/{dataset_id}/refreshes"
        )
        schema = [
            ("refreshType", "RefreshType", "string"),
            ("status", "Status", "string"),
            (
                lambda df: (df["endTime"] - df["startTime"]) / pd.Timedelta(seconds=1),
                "Seconds",
                "long",
            ),
            ("startTime", "StartTime", "timestamp"),
            ("endTime", "EndTime", "timestamp"),
            ("serviceExceptionJson", "Error", "string"),
            ("requestId", "RequestId", "string"),
            ("id", "RefreshId", "long"),
            ("refreshAttempts", "RefreshAttempts", "string"),
        ]
        api_call = requests.get(url=api_url, headers=self.api_header)
        if api_call.status_code == 200:
            return SparkPandasDataFrame(
                api_call.json()["value"],
                schema,
                indexing_columns="id",
                local_timezone_name=self.local_timezone_name,
            )
        self._raise_api_error(
            "Failed to fetch refresh history!",
            api_call,
            workspace_id=workspace_id,
            dataset_id=dataset_id,
            ignore_unauthorized=ignore_unauthorized,
        )
        return None

    def _get_refresh_history_details(
        self,
        workspace_id: str,
        dataset_id: str,
        request_id: str,
        ignore_unauthorized: bool = False,
    ) -> Optional[SparkPandasDataFrame]:
        """
        Returns the PowerBI dataset refresh history details for the specified refresh.

        :param str workspace_id: the id of the workspace
        :param str dataset_id: the id of the dataset
        :param str request_id: the id of the refresh request
        :param bool ignore_unauthorized: return None on HTTP 401 Unauthorized
        :return: data frame with the refresh history details if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: SparkPandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        api_url = (
            f"{self.powerbi_url}groups/{workspace_id}"
            f"/datasets/{dataset_id}/refreshes/{request_id}"
        )
        schema = [
            ("table", "TableName", "string"),
            ("partition", "PartitionName", "string"),
            ("status", "Status", "string"),
        ]
        api_call = requests.get(url=api_url, headers=self.api_header)
        if api_call.status_code == 200:
            return SparkPandasDataFrame(
                api_call.json()["objects"],
                schema,
                indexing_columns=["table", "partition"],
                sorting_columns=["table", "partition"],
            )
        self._raise_api_error(
            "Failed to fetch refresh history details!",
            api_call,
            workspace_id=workspace_id,
            dataset_id=dataset_id,
            request_id=request_id,
            ignore_unauthorized=ignore_unauthorized,
        )
        return None

    def _get_partition_tables(
        self,
        workspace_id: str,
        dataset_id: str,
        ignore_unauthorized: bool = False,
    ) -> Optional[SparkPandasDataFrame]:
        """
        Returns the PowerBI dataset partition info with table names
        and their refresh times.

        :param str workspace_id: the id of the workspace
        :param str dataset_id: the id of the dataset
        :param bool ignore_unauthorized: return None on HTTP 401 Unauthorized
        :return: data frame with the partition info if succeeded or None if failed
        (when ignore_errors==True)
        :rtype: Pandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        api_url = (
            f"{self.powerbi_url}groups/{workspace_id}"
            f"/datasets/{dataset_id}/executeQueries"
        )
        dax_query = {
            "queries": [{"query": "EVALUATE INFO.PARTITIONS()"}],
            "serializerSettings": {"includeNulls": True},
        }
        schema = [
            ("[Name]", "TableName", "string"),
            (
                "[RefreshedTime]",
                "RefreshTime",
                "timestamp",
            ),
            ("[ModifiedTime]", "ModifiedTime", "timestamp"),
            ("[Description]", "Description", "string"),
            ("[ErrorMessage]", "ErrorMessage", "string"),
            ("[ID]", "Id", "long"),
            ("[TableID]", "TableId", "long"),
            ("[DataSourceID]", "DataSourceId", "int"),
            ("[QueryDefinition]", "QueryDefinition", "string"),
            ("[State]", "State", "int"),
            ("[Type]", "Type", "int"),
            ("[PartitionStorageID]", "PartitionStorageId", "int"),
            ("[Mode]", "Mode", "int"),
            ("[DataView]", "DataView", "int"),
            ("[SystemFlags]", "SystemFlags", "int"),
            (
                "[RetainDataTillForceCalculate]",
                "RetainDataTillForceCalculate",
                "boolean",
            ),
            ("[RangeStart]", "RangeStart", "timestamp"),
            ("[RangeEnd]", "RangeEnd", "timestamp"),
            ("[RangeGranularity]", "RangeGranularity", "int"),
            ("[RefreshBookmark]", "RefreshBookmark", "string"),
            ("[QueryGroupID]", "QueryGroupId", "int"),
            ("[ExpressionSourceID]", "ExpressionSourceId", "int"),
            ("[MAttributes]", "MAttributes", "int"),
        ]
        api_call = requests.post(url=api_url, headers=self.api_header, json=dax_query)
        if api_call.status_code == 200:
            return SparkPandasDataFrame(
                api_call.json()["results"][0]["tables"][0]["rows"],
                schema,
                indexing_columns="TableId",
                sorting_columns="TableName",
                local_timezone_name=self.local_timezone_name,
            )
        self._raise_api_error(
            "Failed to fetch partition info!",
            api_call,
            post_body=str(dax_query),
            workspace_id=workspace_id,
            dataset_id=dataset_id,
            ignore_unauthorized=ignore_unauthorized,
        )
        return None

    def _verify_workspace(self, *, force_verify: bool = False) -> bool:
        """
        Gets workspace ID based on the workspace name, or shows all workspaces
            if no parameter was specified.

        :param bool force_verify: True if the workspace_id should be verified as well
        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if self.workspace_id is None or force_verify:
            workspaces = self._get_workspaces()
            if workspaces is None:
                return False
            df = workspaces.get_pandas_df()
            if self.workspace_id is not None:
                if self.workspace_id not in df.WorkspaceId.values:
                    self._raise_error(
                        f"Workspace id '{self.workspace_id}' cannot be found!"
                    )
                    return False
                return True
            if self.workspace_name is None:
                workspaces.show(
                    "Available workspaces:",
                    "No workspaces found.",
                    filter_columns=[
                        ("WorkspaceId", "workspace_id"),
                        ("WorkspaceName", "workspace_name"),
                    ],
                )
                return False
            rows = df.loc[df.WorkspaceName == self.workspace_name, "WorkspaceId"]
            if rows.empty:
                self._raise_error(
                    f"Workspace name '{self.workspace_name}' cannot be found!"
                )
                return False
            self.workspace_id = rows.values[0]

        return True

    def _verify_dataset(self, *, force_verify: bool = False) -> bool:
        """
        Gets dataset ID based on the dataset name, or shows all datasets
            if no parameter was specified.

        :param bool force_verify: True if the dataset_id should be verified as well
        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if self.dataset_id is None or force_verify:
            datasets = self._get_datasets(self.workspace_id)
            if datasets is None:
                return False
            df = datasets.get_pandas_df()
            if df.empty and not self._verify_workspace(force_verify=True):
                return False
            if self.dataset_id is not None:
                if self.dataset_id not in df.DatasetId.values:
                    self._raise_error(
                        f"Dataset id '{self.dataset_id}' cannot be found!"
                    )
                    return False
                return True
            if self.dataset_name is None:
                datasets.show(
                    "Available datasets:",
                    "No datasets found.",
                    filter_columns=[
                        ("DatasetId", "dataset_id"),
                        ("DatasetName", "dataset_name"),
                    ],
                )
                return False
            rows = df.loc[df.DatasetName == self.dataset_name, "DatasetId"]
            if rows.empty:
                self._raise_error(
                    f"Dataset name '{self.dataset_name}' cannot be found, "
                    "or the dataset doesn't have a user with the required permissions!"
                )
                return False
            self.dataset_id = rows.values[0]
            if self.table_names is not None and (
                not self.table_names or not isinstance(self.table_names, list)
            ):
                tables = self._get_partition_tables(self.workspace_id, self.dataset_id)
                if tables is not None:
                    tables.show(
                        "Available tables:",
                        "The dataset has not tables.",
                        filter_columns=[("TableName", "table_name")],
                    )
                return False

        return True

    def _connect(self) -> bool:
        """
        Connects or reconnects to the PowerBI API.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if not self._get_access_token():
            return False
        if not self._verify_workspace():
            return False
        if not self._verify_dataset():
            return False

        return True

    def _connect_and_get_workspaces(
        self,
        *,
        skip_read_only: bool = False,
    ) -> Optional[List[Tuple[str, str]]]:
        """
        Connects or reconnects to the PowerBI API and returns a list of
        workspace id's and names.

        :param bool skip_read_only: True to exclude read-only workspaces.
        :rtype: SparkPandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if not self._get_access_token():
            return None
        if (
            self.workspace_id is not None
            or self.workspace_name is not None
            or self.dataset_id is not None
            or self.dataset_name is not None
        ):
            if not self._verify_workspace():
                return None
        if self.dataset_id is not None or self.dataset_name is not None:
            if not self._verify_dataset():
                return None
        workspace_list = [(self.workspace_id, self.workspace_name)]
        if self.workspace_id is None:
            workspaces = self._get_workspaces()
            if workspaces is None:
                return None
            workspace_list = [
                (df.WorkspaceId, df.WorkspaceName)
                for _, df in workspaces.get_pandas_df().iterrows()
                if not (skip_read_only and df.IsReadOnly)
            ]
        return workspace_list

    def _combine_datasets(self) -> Optional[SparkPandasDataFrame]:
        """
        Returns a list of PowerBI datasets for a single workspace or
        combined lists for datasets from all workspaces.

        :return: data frame with PowerBI datasets if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: SparkPandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        workspace_list = self._connect_and_get_workspaces()
        if workspace_list is None:
            return None
        if self.workspace_id is not None:
            return self._get_datasets(self.workspace_id)
        result = None
        for workspace_id, workspace_name in workspace_list:
            datasets = self._get_datasets(workspace_id)
            if datasets is None:
                return None
            prefix_columns = [("WorkspaceName", workspace_name, "string")]
            suffix_columns = [("WorkspaceId", workspace_id, "string")]
            result = datasets.append(result, prefix_columns, suffix_columns)

        return result

    def _combine_dataframes(
        self,
        function: Callable[[str, str, bool], Optional[SparkPandasDataFrame]],
        *,
        skip_read_only: bool = False,
        skip_not_refreshable: bool = False,
        skip_effective_identity: bool = False,
    ) -> Optional[SparkPandasDataFrame]:
        """
        Returns a single data frame loaded from PowerBI, or the same
        combined data frame fetched from all workspaces and datasets.

        :param callable function: The function that returns the data frame
        :param bool skip_read_only: True to exclude read-only workspaces.
        :param bool skip_not_refreshable: True to exclude PowerBI datasets
           that are not refreshable.
        :param bool skip_effective_identity: True to exclude PowerBI datasets
            requiring an effective identity
            (effective identity is not supported in this version of spetlr).
        :return: requested data frame if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: SparkPandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        workspace_list = self._connect_and_get_workspaces(skip_read_only=skip_read_only)
        if workspace_list is None:
            return None
        if self.workspace_id is not None and self.dataset_id is not None:
            return function(self.workspace_id, self.dataset_id)
        result = None
        for workspace_id, workspace_name in workspace_list:
            datasets = self._get_datasets(workspace_id)
            if datasets is None:
                return None
            for dataset_id, dataset_name in (
                (df.DatasetId, df.DatasetName)
                for _, df in datasets.get_pandas_df().iterrows()
                if not (skip_not_refreshable and not df.IsRefreshable)
                and not (
                    skip_effective_identity
                    and (
                        df.IsEffectiveIdentityRequired
                        or df.IsEffectiveIdentityRolesRequired
                    )
                )
                and (df.ConfiguredBy.lower() if df.ConfiguredBy else None)
                not in self.exclude_creators
            ):
                data = function(workspace_id, dataset_id, True)
                if data is None:
                    continue
                prefix_columns = [
                    ("WorkspaceName", workspace_name, "string"),
                    ("DatasetName", dataset_name, "string"),
                ]
                suffix_columns = [
                    ("WorkspaceId", workspace_id, "string"),
                    ("DatasetId", dataset_id, "string"),
                ]
                if self.workspace_id is not None:
                    prefix_columns = prefix_columns[1:]
                    suffix_columns = suffix_columns[1:]
                result = data.append(result, prefix_columns, suffix_columns)

        return result

    def _combine_refresh_history_details(self):
        """
        Returns a single data frame with refresh history details, or the same
        combined data frame fetched from all workspaces and datasets.
        Only "ViaEnhancedApi" and "Completed" requests are included
        (i.e. when the refresh was executed with the "table_names"
        parameter specified).

        :return: data frame with refresh history details if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: SparkPandas data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        history = self._combine_dataframes(
            self._get_refresh_history, skip_read_only=True, skip_not_refreshable=True
        )
        if history is None:
            return None
        result = None
        for workspace_id, workspace_name, dataset_id, dataset_name, request_id in (
            (
                self.workspace_id if self.workspace_id else df.WorkspaceId,
                self.workspace_name if self.workspace_id else df.WorkspaceName,
                self.dataset_id if self.dataset_id else df.DatasetId,
                self.dataset_name if self.dataset_id else df.DatasetName,
                df.RequestId,
            )
            for _, df in history.get_pandas_df().iterrows()
            if df.RefreshType == "ViaEnhancedApi" and df.Status == "Completed"
        ):
            data = self._get_refresh_history_details(
                workspace_id, dataset_id, request_id, True
            )
            if data is None:
                continue
            prefix_columns = [
                ("WorkspaceName", workspace_name, "string"),
                ("DatasetName", dataset_name, "string"),
            ]
            suffix_columns = [
                ("WorkspaceId", workspace_id, "string"),
                ("DatasetId", dataset_id, "string"),
                ("RequestId", request_id, "string"),
            ]
            if self.dataset_id is not None:
                prefix_columns = prefix_columns[2:]
                suffix_columns = suffix_columns[2:]
            elif self.workspace_id is not None:
                prefix_columns = prefix_columns[1:]
                suffix_columns = suffix_columns[1:]
            result = data.append(result, prefix_columns, suffix_columns)

        return result

    def _get_last_refresh(self, *, deep_check: bool = False) -> bool:
        """
        Gets the latest record in the PowerBI dataset refresh history.

        :param bool deep_check: Used only with check() - ignore history rows
            where refresh is in progress or the refresh type doesn't match
        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        self.last_status = None
        self.last_exception = None
        self.is_enhanced = False
        self.last_refresh_utc = None
        self.table_name = None
        if not self._connect():
            return False
        history = self._get_refresh_history(self.workspace_id, self.dataset_id)
        if history is None:
            return False
        df = history.get_pandas_df()
        if not df.empty:
            first = 0
            while deep_check and (
                df.RefreshType.iloc[first] == "ViaEnhancedApi"
                and df.shape[0] > first + 1
                and (
                    self.table_names is None
                    or (first == 0 and df.Status.iloc[0] == "Unknown")
                )
            ):
                first += 1
            self.last_status = df.Status.iloc[first]
            self.last_exception = df.Error.iloc[first]
            self.is_enhanced = df.RefreshType.iloc[first] == "ViaEnhancedApi"
            # calculate the average duration of all previous API refresh calls
            # when there were no table names specified
            mean = df.loc[
                (df.RefreshType == "ViaApi") & (df.Status == "Completed"),
                df.Seconds.name,
            ].mean()
            if pd.isna(mean) or self.table_names:
                self.last_duration_in_seconds = 0
            else:
                self.last_duration_in_seconds = int(mean)
            if self.last_status == "Completed":
                refresh_time = df["EndTime" + history.time_column_suffix].iloc[first]
                if self.table_names is not None:
                    tables = self._get_partition_tables(
                        self.workspace_id, self.dataset_id, ignore_unauthorized=True
                    )
                    if tables is not None:
                        df = tables.get_pandas_df()
                        df = df[
                            df.TableName.isin(self.table_names)
                            & df.ErrorMessage.notnull()
                        ]
                        if not df.empty:
                            self.last_status = "Failed"
                            self.last_exception = df.ErrorMessage.iloc[0]
                            self.table_name = df.TableName.iloc[0]
                            return True
                        else:
                            df = tables.get_pandas_df()
                            df = df[df.TableName.isin(self.table_names)]
                            if not df.empty:
                                column = "RefreshTime" + tables.time_column_suffix
                                idx = df[column].idxmin()
                                self.table_name = df.TableName.loc[idx]
                                refresh_time = df[column].loc[idx]

                self.last_refresh_utc = history.get_utc_time(refresh_time)
                self.last_refresh_str = history.get_local_time_str(refresh_time)

        return True

    def _verify_last_refresh(self) -> bool:
        """
        Checks if the last refresh of the PowerBI dataset completed successfully,
            and verifies if it happened recently enough.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if self.last_status is None:
            print("No refresh was triggered yet.")
        elif self.last_status == "Completed":
            if self.last_refresh_utc is None:
                self._raise_error("Completed at unknown refresh time!")
            else:
                min_refresh_time_utc = datetime.now(utc) - timedelta(
                    minutes=self.max_minutes_after_last_refresh
                )
                if (
                    self.max_minutes_after_last_refresh > 0
                    and self.last_refresh_utc < min_refresh_time_utc
                ):
                    self._raise_error(
                        (
                            "Last refresh finished more than "
                            f"{self.max_minutes_after_last_refresh} "
                            f"minutes ago at {self.last_refresh_str} !"
                        )
                        if self.table_name is None
                        else (
                            f"Last refresh of the table '{self.table_name}' finished "
                            f"more than {self.max_minutes_after_last_refresh} "
                            f"minutes ago at {self.last_refresh_str} !"
                        )
                    )
                else:
                    print(
                        f"Refresh completed successfully at {self.last_refresh_str}."
                        if self.table_name is None
                        else (
                            f"Last refresh of '{self.table_name}' (the oldest in "
                            "the specified tables) completed successfully "
                            f"at {self.last_refresh_str}."
                        )
                    )
                    return True
        elif self.last_status == "Unknown":
            self._raise_error("Refresh is still in progress!")
        elif self.last_status in ["Canceled", "Cancelled"]:
            self._raise_error("Refresh has been canceled!")
        elif self.last_status == "Disabled":
            self._raise_error("Refresh is disabled!")
        elif self.last_status == "Failed":
            self._raise_error(
                f"Last refresh failed! {self.last_exception}"
                if self.table_name is None
                else f"Last refresh for the table '{self.table_name}' "
                f"failed! {self.last_exception}"
            )
        else:
            self._raise_error(
                f"Unknown refresh status: {self.last_status}! {self.last_exception}"
            )
        return False

    def _get_refresh_argument_json(
        self, with_wait: bool
    ) -> Optional[Dict[str, object]]:
        """
        Returns the HTTP body of the PowerBI refresh API call
        containing table names to refresh or other parameters if necessary.

        :param bool with_wait: True if we need to wait for the refresh to finish.
        :rtype: JSON object or None
        """

        result = {}
        if self.table_names:
            result = {
                "type": "full",
                "commitMode": "transactional",
                "objects": [{"table": table} for table in self.table_names],
                "applyRefreshPolicy": "false",
            }
        elif self.mail_on_failure or self.mail_on_completion:
            result["notifyOption"] = (
                "MailOnCompletion" if self.mail_on_completion else "MailOnFailure"
            )
        if self.number_of_retries > 0 and not with_wait:
            if self.table_names:
                result["retryCount"] = self.number_of_retries
            else:
                print(
                    "The 'number_of_retries' parameter is ignored in "
                    "start_refresh() if 'table_names' is not specified!"
                )
        if self.max_parallelism is not None:
            result["maxParallelism"] = self.max_parallelism
        if self.apply_refresh_policy is not None:
            result["applyRefreshPolicy"] = str(self.apply_refresh_policy).lower()
        if self.dataset_refresh_type is not None:
            result["type"] = self.dataset_refresh_type
        if self.dataset_commit_mode is not None:
            result["commitMode"] = self.dataset_commit_mode
        if self.incremental_effective_date is not None:
            result["effectiveDate"] = self.incremental_effective_date

        return result if result else None

    def _trigger_new_refresh(self, *, with_wait: bool = True) -> bool:
        """
        Starts a refresh of the PowerBI dataset.

        :param bool with_wait: True if we need to wait for the refresh to finish.
        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if self.last_status == "Disabled":
            self._raise_error("Refresh is disabled!")
        elif self.last_status == "Unknown":
            if self.is_enhanced or not self.table_names:
                if with_wait:
                    print("Refresh is in progress, waiting until completed.")
                else:
                    self._raise_error("Refresh is already in progress!")
                return True
            print("Refresh of the whole dataset is in progress. Nothing to do.")
        else:
            if self.last_status == "Failed":
                print(f"Warning: Last refresh failed! {self.last_exception}")
                print()
            elif self.last_status is not None and self.last_status not in [
                "Completed",
                "Canceled",
                "Cancelled",
            ]:
                print(
                    f"Unknown refresh status: {self.last_status}! {self.last_exception}"
                )
                print()

            post_body = self._get_refresh_argument_json(with_wait)
            api_url = (
                f"{self.powerbi_url}groups/{self.workspace_id}"
                f"/datasets/{self.dataset_id}/refreshes"
            )
            api_call = requests.post(
                url=api_url, headers=self.api_header, json=post_body
            )
            if api_call.status_code == 202:
                print("A new refresh has been successfully triggered.")
                self.last_status = "Unknown"
                return True
            else:
                self._raise_api_error(
                    "Failed to trigger a refresh!",
                    api_call,
                    post_body=str(post_body),
                    workspace_id=self.workspace_id,
                    dataset_id=self.dataset_id,
                )

        return False

    def _get_seconds_to_wait(self, elapsed_seconds: int) -> int:
        """
        Returns the number of seconds to wait before rechecking
            if the refresh has completed. The method makes sure as few requests
            to the PowerBI API as possible would be made.

        :param int elapsed_seconds: The number of seconds elapsed so far.
        :return: number of seconds to wait
        :rtype: int
        """

        # Microsoft set a limit of how many API requests in an hour can be made
        # to the PowerBI API. This logic is a fine compromise between how often
        # (small delay) we check through polling if the refresh has finished and
        # how soon (large delay) will the refresh() method complete.

        remaining_seconds = self.timeout_in_seconds - elapsed_seconds

        wait_seconds = elapsed_seconds
        if self.last_duration_in_seconds > 0:
            wait_seconds = abs(self.last_duration_in_seconds - elapsed_seconds)

        wait_seconds = (
            15 if wait_seconds < 40 else 60 if wait_seconds < (3 * 60) else (5 * 60)
        )

        return (
            remaining_seconds
            if (0 < remaining_seconds < wait_seconds)
            else wait_seconds
        )

    def check(self) -> bool:
        """
        Checks if the last refresh of the PowerBI dataset or its selected tables
            completed successfully, and verifies if it happened recently enough.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        return self._get_last_refresh(deep_check=True) and self._verify_last_refresh()

    def start_refresh(self) -> bool:
        """
        Starts a refresh of the PowerBI dataset without waiting.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        return self._get_last_refresh() and self._trigger_new_refresh(with_wait=False)

    def refresh(self) -> bool:
        """
        Starts a refresh of the PowerBI dataset and waits until completed.

        :return: True if succeeded or False if failed (when ignore_errors==True)
        :rtype: bool
        :raises PowerBiException: if failed and ignore_errors==False
        """

        retries = self.number_of_retries
        start_time = time.time()
        if not self._get_last_refresh():
            return False
        restart = self.last_status == "Unknown" and self.is_enhanced
        if not self._trigger_new_refresh():
            return False

        while True:
            elapsed_seconds = int(time.time() - start_time)
            if elapsed_seconds >= self.timeout_in_seconds:
                print("Timeout!")
                break
            wait_seconds = self._get_seconds_to_wait(elapsed_seconds)
            print(f"Waiting {wait_seconds} seconds...")
            time.sleep(wait_seconds)

            if not self._get_last_refresh():
                return False

            if (self.last_status == "Failed" and retries > 0) or (
                self.last_status == "Completed" and restart
            ):
                if not restart:
                    retries = retries - 1
                restart = False
                if not self._trigger_new_refresh():
                    return False
                continue

            if self.last_status != "Unknown":
                break

        return self._verify_last_refresh()

    def show_history(self) -> None:
        """
        Displays the refresh history of a PowerBI dataset.

        :rtype: None
        :raises PowerBiException: if failed and ignore_errors==False
        """

        history = self._combine_dataframes(
            self._get_refresh_history, skip_read_only=True, skip_not_refreshable=True
        )
        if history is not None:
            history.show("Refresh history:", "The refresh history is empty.")

    def get_history(self) -> Optional[DataFrame]:
        """
        Returns the refresh history of a PowerBI dataset in a Spark data frame.

        :return: the data frame with the refresh history if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: Spark data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        history = self._combine_dataframes(
            self._get_refresh_history, skip_read_only=True, skip_not_refreshable=True
        )
        return history.get_spark_df() if history is not None else None

    def show_history_details(self) -> None:
        """
        Displays refresh history details of a PowerBI dataset.

        :rtype: None
        :raises PowerBiException: if failed and ignore_errors==False
        """

        details = self._combine_refresh_history_details()
        if details is not None:
            details.show("Refresh history details:", "The refresh history is empty.")

    def get_history_details(self) -> Optional[DataFrame]:
        """
        Returns refresh history details of a PowerBI dataset in a Spark data frame.

        :return: the data frame with the refresh history details if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: Spark data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        history = self._combine_refresh_history_details()
        return history.get_spark_df() if history is not None else None

    def show_tables(self) -> None:
        """
        Displays the tables of a PowerBI dataset.

        :rtype: None
        :raises PowerBiException: if failed and ignore_errors==False
        """

        tables = self._combine_dataframes(
            self._get_partition_tables, skip_effective_identity=True
        )
        if tables is not None:
            tables.show("Dataset tables:", "No dataset tables found.")

    def get_tables(self) -> Optional[DataFrame]:
        """
        Returns the tables of a PowerBI dataset as a Spark data frame.

        :return: the data frame with the partition tables if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: Spark data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        tables = self._combine_dataframes(
            self._get_partition_tables, skip_effective_identity=True
        )
        return tables.get_spark_df() if tables is not None else None

    def show_workspaces(self) -> None:
        """
        Displays the available workspaces.

        :rtype: None
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if self._get_access_token():
            workspaces = self._get_workspaces()
            if workspaces is not None:
                workspaces.show("Workspaces:", "No workspaces found.")

    def get_workspaces(self) -> Optional[DataFrame]:
        """
        Returns the available workspaces as a Spark data frame.

        :return: the data frame with the workspaces if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: Spark data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        if self._get_access_token():
            workspaces = self._get_workspaces()
            return workspaces.get_spark_df() if workspaces is not None else None
        return None

    def show_datasets(self) -> None:
        """
        Displays the available datasets.

        :rtype: None
        :raises PowerBiException: if failed and ignore_errors==False
        """

        datasets = self._combine_datasets()
        if datasets is not None:
            datasets.show("Datasets:", "No datasets found.")

    def get_datasets(self) -> Optional[DataFrame]:
        """
        Returns the available datasets as a Spark data frame.

        :return: the data frame with the datasets if succeeded,
            or None if failed (when ignore_errors==True)
        :rtype: Spark data frame
        :raises PowerBiException: if failed and ignore_errors==False
        """

        datasets = self._combine_datasets()
        return datasets.get_spark_df() if datasets is not None else None



from .eh2bronze.EhToDeltaBronzeOrchestrator import (  # noqa: F401
    EhToDeltaBronzeOrchestrator,
)
from .eh2silver.EhToDeltaSilverOrchestrator import (  # noqa: F401
    EhToDeltaSilverOrchestrator,
)
from .ehjson2delta.EhJsonToDeltaOrchestrator import (  # noqa: F401
    EhJsonToDeltaOrchestrator,
)



import json
import os
from datetime import datetime
from typing import Union

from deprecated import deprecated
from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from pyspark.sql.utils import AnalysisException

from spetlr import dbg
from spetlr.configurator.configurator import Configurator
from spetlr.eh.eh_exceptions import SpetlrEhInitException, SpetlrEhLogicException
from spetlr.eh.PartitionSpec import PartitionSpec
from spetlr.functions import init_dbutils
from spetlr.spark import Spark


@deprecated(
    reason="Use EventHubCaptureExtractor instead.",
)
class EventHubCapture:
    """Class to access eventhub capture files as table.

    The constructor argument should be given as a key that can be found in
    the Table Configurator, where the relevant table must be marked with
    format=avro and with a partitioning from the set of known partitions: y,m,d,h
    """

    max_partition: Union[PartitionSpec, None]
    name: str
    path: str
    format: str
    partitioning: str
    auto_create: bool

    @classmethod
    def from_tc(cls, id: str):
        tc = Configurator()
        return cls(
            name=tc.get(id, "name"),
            path=tc.get(id, "path"),
            format=tc.get(id, "format"),
            partitioning=tc.get(id, "partitioning"),
        )

    def __init__(
        self,
        name: str,
        path: str,
        format: str,
        partitioning: str,
        auto_create: bool = True,
    ):
        self.name = name
        self.path = path
        self.format = format
        self.partitioning = partitioning.lower()
        self.auto_create = auto_create

        assert self.format == "avro"
        self._validate_partitioning()
        self.max_partition = None

    def _validate_partitioning(self, partitioning: str = None):
        if not partitioning:
            partitioning = self.partitioning

        if partitioning not in ["ymd", "ymdh"]:
            raise ValueError("partitioning must be ymd or ymdh")

        return True

    def _create_table(self):
        """Create the external avro table"""
        avro_schema = {
            "namespace": "org.apache.hive",
            "name": "first_schema",
            "type": "record",
            "fields": [
                {"name": "SequenceNumber", "type": "long"},
                {"name": "Offset", "type": "string"},
                {"name": "EnqueuedTimeUtc", "type": "string"},
                {"name": "Body", "type": "string"},
            ],
        }
        Spark.get().sql(
            f"""
            CREATE TABLE {self.name}
            ({", ".join(f"{c} STRING" for c in self.partitioning)})
            STORED AS AVRO
            PARTITIONED BY ({",".join(self.partitioning)})
            LOCATION "{self.path}"
            TBLPROPERTIES ('avro.schema.literal'='{json.dumps(avro_schema)}');
        """
        )

    def _discover_first_partition(self) -> PartitionSpec:
        """Add the first partition by discovering it from filesystem."""
        # raise if the table already has partitions
        if Spark.get().sql(f"SHOW PARTITIONS {self.name}").count():
            raise SpetlrEhLogicException("Table partitions already initialized.")

        dbutils = init_dbutils()
        partition = PartitionSpec()

        # discover each part
        for c in self.partitioning:
            # loop over items in path
            full_path = self.path + "/" + partition.as_path()
            for file_info in dbutils.fs.ls(full_path):
                if not (
                    file_info.name.startswith(c + "=") and file_info.name.endswith("/")
                ):
                    continue
                extraction = int(file_info.name[2:-1])
                previous = partition.__getattribute__(c)
                # keep only the lowest
                if previous is None or extraction < previous:
                    partition.__setattr__(c, extraction)

            # done looping over items.
            if partition.__getattribute__(c) is None:
                # There is probably no data, yet.
                raise SpetlrEhInitException(
                    f"unable to discover first partition at '{full_path}'"
                )

        # we have discovered all parts of the partition.
        Spark.get().sql(
            f"""ALTER TABLE {self.name} ADD PARTITION ({partition.as_sql_spec()})"""
        )

        # return the partition in path form so that the repair can continue
        return partition

    def _get_max_partition(self) -> PartitionSpec:
        """get the current highest partition either
        1. from memory
        2. read the partition table
        3. initialize from disk
        """
        if self.max_partition is None:
            # memory access failed.
            # try reading partition table
            try:
                max_p = (
                    Spark.get()
                    .sql(f"SHOW PARTITIONS {self.name}")
                    .groupBy()
                    .agg(f.max("partition"))
                    .collect()
                )[0][0]
                if max_p is not None:
                    dbg(f"There is an old partition! {max_p}")
                    self.max_partition = PartitionSpec.from_path(max_p)
                # value will be None if there is no value
            except AnalysisException:
                # no such table. Better create one
                if self.auto_create:
                    self._create_table()
                else:
                    raise SpetlrEhInitException("external table does not exist")

        if self.max_partition is None:
            # reading partition metadata failed.
            # initialize first partition from disk.
            self.max_partition = self._discover_first_partition()

        return self.max_partition

    def _repair_partitioning(self):
        self._get_max_partition()

        # compare with current wall-clock to see if partitions are missing
        now = datetime.now()

        additions = []
        while self.max_partition.is_earlier_than_dt(now):
            # we know that the partition table is behind
            next_partition = self.max_partition.next()

            dbg(
                "we are behind on the partitions. "
                f"Try adding next {next_partition.as_path()}"
            )

            # check for existence of partition. If it does not exits,
            # databricks would otherwise try to create the folder which breaks
            # for read-only mounts.
            try:
                # Here we do not use dbutils because it does not have a meaningful
                # exception when the file does not exist. Also, it seems slower.
                # The only negative consequence is that the partitions will not
                # be updated over databricks-connect
                os.listdir("/dbfs" + self.path + "/" + next_partition.as_path())
                additions.append(f"PARTITION ({next_partition.as_sql_spec()})")

            except FileNotFoundError:
                pass

            self.max_partition = next_partition

        while additions:
            # We need to add partitions
            parts_per_batch = min(len(additions), 24)
            dbg(f"adding {parts_per_batch} partitions")

            # add partitions in batch to prevent an sql line with hundreds of parts
            batch = [additions.pop() for _ in range(parts_per_batch)]
            Spark.get().sql(f"ALTER TABLE {self.name} ADD " + " ".join(batch))
        else:
            dbg("no partitions to add")

    def read(self) -> DataFrame:
        self._repair_partitioning()
        return (
            Spark.get()
            .table(self.name)
            .withColumn(
                "pdate",
                f.expr(
                    'make_timestamp(y,m,d,h,0,0,"UTC")'
                    if "h" in self.partitioning
                    else 'make_timestamp(y,m,d,0,0,0,"UTC")'
                ),
            )
        )



import unittest

from spetlr.sql import SqlServer


class DeliverySqlServerTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls) -> None:
        pass

    def test_from_connection_string(self):
        server = "cltest.database.windows.net"
        port = "1433"
        database = "testdatabase!@_"
        user = "testuser"
        password = "testpwd"

        connection_string = f"""
        Server=tcp:{server},{port};
        Database={database};
        User ID={user};
        Password={password};
        Encrypt=true;
        Connection Timeout=30;"""

        test_server = SqlServer(connection_string=connection_string)
        sql_server = SqlServer(server, database, user, password, port)

        self.assertEqual(test_server.odbc, sql_server.odbc)
        self.assertEqual(test_server.url, sql_server.url)

    def test_from_connection_string2(self):
        server = "cltest.database.windows.net"
        port = "1433"
        database = "test.database!@"
        user = "test_user1"
        password = "_@b3&yy"

        connection_string = f"""
        Addr={server};
        Initial Catalog={database};
        UID={user};
        PWD={password};
        Encrypt=true;
        Connection Timeout=30;"""

        test_server = SqlServer(connection_string=connection_string)
        sql_server = SqlServer(server, database, user, password, port)

        self.assertEqual(test_server.odbc, sql_server.odbc)
        self.assertEqual(test_server.url, sql_server.url)

    def test_from_connection_string_raises(self):
        connection_string = "Not data=tcp:, not pass, not user"

        self.assertRaises(
            ValueError, SqlServer, None, None, None, None, None, connection_string
        )

    def test_not_enough_params_raises(self):
        self.assertRaises(
            ValueError,
            SqlServer,
            "test",
            None,
            "something missing",
            "nothere",
            "404",
            None,
        )

    def test_spn_url(self):
        server = "cltest.database.windows.net"
        database = "testdatabase!@_"
        spnid = "spnuser"
        spnpassword = "spnpwd"
        timeout = 180

        sql_server = SqlServer(server, database, spnid=spnid, spnpassword=spnpassword)

        expected_jdbc_url = (
            f"jdbc:sqlserver://{server}:1433;"
            f"database={database};"
            f"queryTimeout=0;"
            f"loginTimeout={timeout};"
            "driver=com.microsoft.sqlserver.jdbc.SQLServerDriver;"
            f"AADSecurePrincipalId={spnid};"
            f"AADSecurePrincipalSecret={spnpassword};"
            f"encrypt=true;"
            f"trustServerCertificate=false;"
            f"hostNameInCertificate=*.database.windows.net;"
            f"authentication=ActiveDirectoryServicePrincipal"
        )

        expected_odbc_url = (
            "DRIVER={ODBC Driver 18 for SQL Server};"
            f"SERVER={server};"
            f"DATABASE={database};"
            f"PORT=1433;"
            f"UID={spnid};"
            f"PWD={spnpassword};"
            f"Connection Timeout={timeout}"
            f";Authentication=ActiveDirectoryServicePrincipal"
        )

        self.assertEqual(sql_server.url, expected_jdbc_url)
        self.assertEqual(sql_server.odbc, expected_odbc_url)

    def test_sqluseradmin_url(self):
        server = "cltest.database.windows.net"
        database = "testdatabase!@_"
        user = "user"
        password = "pass"
        timeout = 180

        sql_server = SqlServer(server, database, username=user, password=password)

        expected_jdbc_url = (
            f"jdbc:sqlserver://{server}:1433;"
            f"database={database};"
            f"queryTimeout=0;"
            f"loginTimeout={timeout};"
            "driver=com.microsoft.sqlserver.jdbc.SQLServerDriver;"
            f"user={user};password={password}"
        )

        expected_odbc_url = (
            "DRIVER={ODBC Driver 18 for SQL Server};"
            f"SERVER={server};"
            f"DATABASE={database};"
            f"PORT=1433;"
            f"UID={user};"
            f"PWD={password};"
            f"Connection Timeout={timeout}"
        )

        self.assertEqual(sql_server.url, expected_jdbc_url)
        self.assertEqual(sql_server.odbc, expected_odbc_url)

    def test_both_spn_and_sqluser(self):
        server = "cltest.database.windows.net"
        database = "testdatabase!@_"
        user = "user"
        password = "password"
        spnid = "spnuser"
        spnpassword = "spnpwd"
        exptected_fail = "Use either SPN or SQL user - never both"

        with self.assertRaises(ValueError) as ctx:
            SqlServer(
                server,
                database,
                username=user,
                password=password,
                spnid=spnid,
                spnpassword=spnpassword,
            )
        self.assertEqual(exptected_fail, str(ctx.exception))

        with self.assertRaises(ValueError) as ctx:
            SqlServer(
                server,
                database,
                username=user,
                spnpassword=spnpassword,
            )
        self.assertEqual(exptected_fail, str(ctx.exception))

        with self.assertRaises(ValueError) as ctx:
            SqlServer(
                server,
                database,
                password=password,
                spnid=spnid,
            )
        self.assertEqual(exptected_fail, str(ctx.exception))



import unittest

from spetlr.exceptions import OnlyUseInSpetlrDebugMode
from spetlr.testutils import CleanupTestDatabases


class CleanUpDatabasesAndTablesUnitTest(unittest.TestCase):
    def test_01_debug_databases(self):
        with self.assertRaises(OnlyUseInSpetlrDebugMode):
            CleanupTestDatabases()



import json
from dataclasses import asdict, dataclass
from typing import Dict, Optional

from deprecated import deprecated
from pyspark.sql.utils import AnalysisException

from spetlr import Configurator
from spetlr.configurator.sql.parse_sql import parse_single_sql_statement
from spetlr.deltaspec.exceptions import InvalidSpecificationError
from spetlr.deltaspec.helpers import ensureStr, standard_databricks_location
from spetlr.exceptions import NoSuchValueException
from spetlr.spark import Spark

# TODO: for writing remove checks on comment equality


@deprecated("Class untested in current version of spetlr")
@dataclass
class DeltaDatabaseSpec:
    name: str
    comment: Optional[str] = None
    location: Optional[str] = None
    dbproperties: Dict[str, str] = None

    def __post_init__(self):
        self.name = ensureStr(self.name)
        self.location = standard_databricks_location(self.location)

    def __repr__(self):
        """returns python code that can construct the present object such as
        DeltaDatabaseSpec(
            name="mydb_name",
            comment="cool data",
            location="dbfs:/my/path"
        )
        """
        dbproperties_part = ""
        if self.dbproperties:
            description = ", ".join(
                f'"{k}":"{v}"' for k, v in self.dbproperties.items()
            )
            dbproperties_part = f"dbproperties={{{description}}}, "

        return (
            ", ".join(
                part
                for part in [
                    f"DeltaDatabaseSpec(name={repr(self.name)}",
                    (f"comment={repr(self.comment)}" if self.comment else ""),
                    (f"location={repr(self.location)}" if self.location else ""),
                    dbproperties_part,
                ]
                if part
            )
            + ")"
        )

    @classmethod
    def from_tc(cls, id: str) -> "DeltaDatabaseSpec":
        """Build a DeltaDatabaseSpec instance from what is in the Configurator.
        This may have previously been parsed from sql."""
        c = Configurator()
        try:
            name = c.get(id, "name")
        except NoSuchValueException:
            raise InvalidSpecificationError()

        location = c.get(id, "path", default=None)
        comment = c.get(id, "comment", default=None)
        dbproperties = c.get(id, "dbproperties", default=None)

        return cls(
            name=name, location=location, comment=comment, dbproperties=dbproperties
        )

    @classmethod
    def from_sql(cls, sql: str) -> "DeltaDatabaseSpec":
        """Parse a sigle CREATE DATABASE statement
        to initialize a DeltaDatabaseSpec object."""
        details = parse_single_sql_statement(sql)
        format = details.get("format").lower()
        if format != "db":
            raise InvalidSpecificationError(
                "The sql code is not a create database statement."
                f"Instead, the provided format was {format}"
            )

        init_args = dict(
            name=details.get("name"),
            location=details.get("path"),
            comment=details.get("comment"),
            dbproperties=details.get("dbproperties", {}),
        )

        init_args = {k: v for k, v in init_args.items() if v}
        return cls(**init_args)

    def get_create_sql(self) -> str:
        """Return the SQL code that will create the database that this
        class describes"""
        name_part = f"CREATE SCHEMA IF NOT EXISTS {self.name}"
        comment_part = f"  COMMENT {json.dumps(self.comment)}" if self.comment else ""
        location_part = (
            f"  LOCATION {json.dumps(self.location)}" if self.location else ""
        )
        dbproperties_part = ""
        if self.dbproperties:
            description = ", ".join(
                f"{k}={json.dumps(v)}" for k, v in self.dbproperties.items()
            )
            dbproperties_part = f"  WITH DBPROPERTIES ({description})"
        return "\n".join(
            part
            for part in [name_part, comment_part, location_part, dbproperties_part]
            if part
        )

    @classmethod
    def from_spark(cls, name: str) -> "DeltaDatabaseSpec":
        try:
            rows = Spark.get().sql(f"DESCRIBE SCHEMA {name}").collect()
        except AnalysisException:
            return None  # No such schema

        comment = location = None
        for row in rows:
            if str(row[0]).lower() == "comment":
                comment = str(row[1])
            elif str(row[0]).lower() == "location":
                location = str(row[1])

        # TODO: parsing of dbproperties currently not supported

        return cls(name=name, location=location, comment=comment)

    def fully_substituted(self) -> "DeltaDatabaseSpec":
        """Return a new DeltaDatabaseSpec
        where name and location have been completed via the Configurator."""
        parts = asdict(self)

        c = Configurator()
        details = c.get_all_details()
        parts["name"] = self.name.format(**details)
        parts["location"] = self.location.format(**details)

        return DeltaDatabaseSpec(**parts)

    def matches_to_spark(self) -> bool:
        """Check if the specification of this Database agrees
        with the specification of the database of the same name
        in the spark catalog."""
        full = self.fully_substituted()
        onstorage = self.from_spark(full.name)
        return full == onstorage

    def create(self):
        Spark.get().sql(self.get_create_sql())

    def drop_cascade(self):
        Spark.get().sql(f"DROP DATABASE {self.name} CASCADE")



import unittest
from unittest.mock import MagicMock, patch

from pyspark.sql import DataFrame, Row
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from spetlr.delta import DeltaHandle
from spetlr.deltaspec import DeltaTableSpecBase
from spetlr.deltaspec.DeltaTableSpec import DeltaTableSpec
from spetlr.deltaspec.DeltaTableSpecDifference import DeltaTableSpecDifference
from spetlr.deltaspec.exceptions import TableSpecNotReadable


class TestDeltaTableSpecMethods(unittest.TestCase):

    def setUp(self):
        """Setup a sample DeltaTableSpec object."""
        self.schema = StructType(
            [
                StructField("id", IntegerType(), True),
                StructField("name", StringType(), True),
            ]
        )

        self.spec = DeltaTableSpec(
            name="test_db.test_table",
            schema=self.schema,
            location="abfss://storage/tables/test_table",
            tblproperties={
                "delta.minReaderVersion": "2",
                "delta.minWriterVersion": "5",
            },
        )

    def test_copy(self):
        """Test that copy() returns an independent object that compares equal to the original."""
        spec_copy = self.spec.copy()

        self.assertEqual(self.spec, spec_copy)
        self.assertIsNot(
            self.spec, spec_copy
        )  # Ensure they are different objects in memory

    def test_get_dh(self):
        """Test that get_dh() returns a valid DeltaHandle with expected values."""
        dh = self.spec.get_dh()

        self.assertIsInstance(dh, DeltaHandle)
        self.assertEqual(
            dh.get_tablename(), self.spec.name
        )  # Ensuring the name is correct
        self.assertEqual(
            dh._location, self.spec.location
        )  # Ensuring the location matches
        self.assertEqual(dh._data_format, "delta")  # Ensuring format is always delta

    def test_compare_to(self):
        """Test that compare_to() returns a DeltaTableSpecDifference object."""

        other_spec = self.spec.copy()

        # Run the method to get the actual result
        result = self.spec.compare_to(other_spec)

        # Ensure the result is a DeltaTableSpecDifference instance
        self.assertIsInstance(result, DeltaTableSpecDifference)

        # Ensure base and target are correctly assigned
        self.assertEqual(result.base.name, other_spec.name)
        self.assertEqual(result.base.schema, other_spec.schema)
        self.assertEqual(result.base.tblproperties, other_spec.tblproperties)
        self.assertEqual(result.base.location, other_spec.location)

        self.assertEqual(result.target.name, self.spec.name)
        self.assertEqual(result.target.schema, self.spec.schema)
        self.assertEqual(result.target.tblproperties, self.spec.tblproperties)
        self.assertEqual(result.target.location, self.spec.location)

    @patch.object(DeltaTableSpec, "compare_to_name")
    def test_compare_to_location(self, mock_compare_to_name):
        """Test that compare_to_location() calls compare_to_name correctly."""

        # Mock return value
        mock_compare_to_name.return_value = "mocked_result"

        # Call the method
        result = self.spec.compare_to_location()

        # Ensure the method returns the expected result
        self.assertEqual(result, "mocked_result")

        # Ensure compare_to_name() was called exactly once
        mock_compare_to_name.assert_called_once()

    @patch.object(DeltaTableSpec, "from_name")
    def test_compare_to_name(self, mock_from_name):
        """Test that compare_to_name() correctly compares with the catalog table."""

        # Mock from_name() to return a copied version of self.spec
        other_spec = self.spec.copy()
        mock_from_name.return_value = other_spec

        # Convert expected mock return value to DeltaTableSpecBase
        expected_base = DeltaTableSpecBase(
            name=other_spec.name,
            schema=other_spec.schema,
            options=other_spec.options,
            partitioned_by=other_spec.partitioned_by,
            cluster_by=other_spec.cluster_by,
            tblproperties=other_spec.tblproperties,
            location=other_spec.location,
            comment=other_spec.comment,
            blankedPropertyKeys=["delta.columnMapping.maxColumnId"],
        )

        expected_target = DeltaTableSpecBase(
            name=self.spec.name,
            schema=self.spec.schema,
            options=self.spec.options,
            partitioned_by=self.spec.partitioned_by,
            cluster_by=self.spec.cluster_by,
            tblproperties=self.spec.tblproperties,
            location=self.spec.location,
            comment=self.spec.comment,
            blankedPropertyKeys=["delta.columnMapping.maxColumnId"],
        )

        # Call the method
        result = self.spec.compare_to_name()

        # Ensure the result is a DeltaTableSpecDifference object
        self.assertIsInstance(result, DeltaTableSpecDifference)

        # Ensure base and target match expectations
        self.assertEqual(
            result.base, expected_base
        )  # Convert mock return to DeltaTableSpecBase
        self.assertEqual(result.target, expected_target)

        # Ensure from_name() was called with the correct table name
        mock_from_name.assert_called_once_with(self.spec.name)

    @patch("spetlr.spark.Spark.get")
    def test_ensure_df_schema(self, mock_spark_get):
        """Test that ensure_df_schema() raises TableSpecNotReadable if schema mismatch occurs."""
        # Create a mock DataFrame with an incompatible schema
        mismatched_schema = StructType(
            [
                StructField("id", IntegerType(), True),
                StructField("email", StringType(), True),  # Different field name
            ]
        )

        mock_df = MagicMock(spec=DataFrame)
        mock_df.schema = mismatched_schema

        with self.assertRaises(TableSpecNotReadable):
            self.spec.ensure_df_schema(mock_df)

    @patch.object(DeltaTableSpec, "compare_to_name")
    def test_is_readable(self, mock_compare_to_name):
        """Test that is_readable() returns the correct boolean result."""

        # Case 1: When is_readable() should return True
        mock_compare_to_name.return_value.is_readable.return_value = True
        result = self.spec.is_readable()
        self.assertTrue(result)

        # Case 2: When is_readable() should return False
        mock_compare_to_name.return_value.is_readable.return_value = False
        result = self.spec.is_readable()
        self.assertFalse(result)

        # Ensure compare_to_name() was called twice (once per case)
        self.assertEqual(mock_compare_to_name.call_count, 2)



from typing import Any, Dict, List, Union

import pyspark.sql.types as T
from pyspark.sql import DataFrame


class TableHandle:
    def read(self) -> DataFrame:
        raise NotImplementedError()

    def overwrite(
        self, df: DataFrame, mergeSchema: bool = None, overwriteSchema: bool = None
    ) -> None:
        raise NotImplementedError()

    def append(self, df: DataFrame, mergeSchema: bool = None) -> None:
        raise NotImplementedError()

    def truncate(self) -> None:
        raise NotImplementedError()

    def drop(self) -> None:
        raise NotImplementedError()

    def drop_and_delete(self) -> None:
        raise NotImplementedError()

    def write_or_append(self, df: DataFrame, mode: str) -> None:
        raise NotImplementedError()

    def upsert(self, df: DataFrame, join_cols: List[str]) -> Union[DataFrame, None]:
        raise NotImplementedError()

    def get_tablename(self) -> str:
        raise NotImplementedError()

    def read_stream(self):
        raise NotImplementedError()
        pass

    def delete_data(
        self, comparison_col: str, comparison_limit: Any, comparison_operator: str
    ) -> None:
        pass

    def get_table_id(self):
        raise NotImplementedError()

    def get_schema(self) -> T.StructType:
        raise NotImplementedError()

    def set_schema(self, schema: T.StructType) -> "TableHandle":
        raise NotImplementedError()

    def get_options(self) -> Dict[str, Any]:
        raise NotImplementedError()

    def set_options(self, options: Dict[str, Any]) -> "TableHandle":
        raise NotImplementedError()



import copy
import dataclasses
import json
from dataclasses import _MISSING_TYPE, asdict, dataclass
from typing import Dict, List, Optional

from deprecated import deprecated
from pyspark.sql.types import StructField, StructType

from spetlr.deltaspec.DeltaTableSpecBase import DeltaTableSpecBase
from spetlr.deltaspec.exceptions import TableSpectNotEnforcable
from spetlr.deltaspec.helpers import TableName
from spetlr.schema_manager import SchemaManager


@deprecated("Class untested in current version of spetlr")
@dataclass
class DeltaTableSpecDifference:
    """This class contains two DeltaTableSpec, base and target.
    The class can make statements about the degree of agreement
    and can generate the ALTER TABLE statements that are
    needed to transform the base into the target.
    As a dataclass, the repr string for this class is able to
    restore the class object.
    """

    base: Optional[DeltaTableSpecBase]
    target: DeltaTableSpecBase

    def __post_init__(self):
        # decouple from the initializing objects
        self.base = (
            DeltaTableSpecBase(**asdict(self.base.copy())) if self.base else None
        )
        self.target = DeltaTableSpecBase(**asdict(self.target.copy()))

        # Only compare table names up to the highest level
        # that both DeltaTableSpec have information about.
        # so a nameless table can still compare equal to
        # another table. Also, unless both tables specify
        # the catalog, the catalog is ignored.

        if self.base:
            b_name = TableName.from_str(self.base.name)
            t_name = TableName.from_str(self.target.name)
            level = min(t_name.level(), b_name.level())
            self.base.name = str(b_name.to_level(level))
            self.target.name = str(t_name.to_level(level))

    def ignore(self, *keys: str):
        """Returns a new DeltaTableSpecDifference where the given set of
        keys are reset to default for both base and target."""

        # map each name to either a default factory object,
        # or a default object, or None
        # depending on what we have
        delta_spec_defaults = {
            f.name: (
                f.default_factory()
                if not isinstance(f.default_factory, _MISSING_TYPE)
                else f.default if not isinstance(f.default, _MISSING_TYPE) else None
            )
            for f in dataclasses.fields(self.target)
        }
        ignored_values = {key: delta_spec_defaults[key] for key in keys}
        new_base = (
            None
            if self.base is None
            else dataclasses.replace(self.base, **ignored_values)
        )
        new_target = dataclasses.replace(self.target, **ignored_values)
        return DeltaTableSpecDifference(base=new_base, target=new_target)

    def nullbase(self) -> bool:
        """is the comparison to a null base. Meaing there is no table."""
        return self.base is None

    def complete_match(self) -> bool:
        """A complete match = no difference"""
        if self.base is None:
            return False
        return self.base == self.target

    def is_different(self) -> bool:
        return not self.complete_match()

    def name_match(self):
        if self.base is None:
            return False
        return self.base.name == self.target.name

    def location_match(self):
        if self.base is None:
            return False
        return self.base.location == self.target.location

    def schema_match(self):
        if self.base is None:
            return False
        return self.base.schema == self.target.schema

    def metadata_match(self):
        if self.base is None:
            return False
        return (
            (self.base.options == self.target.options)
            and (self.base.partitioned_by == self.target.partitioned_by)
            and (self.base.cluster_by == self.target.cluster_by)
            and (self.base.comment == self.target.comment)
            and (self.base.tblproperties == self.target.tblproperties)
        )

    def is_readable(self) -> bool:
        """Returns true if the .read() and .append() methods can be executed
        on the target without throwing exceptions, under the assumption that
        the base represents to true state of data on disk."""
        if self.complete_match():
            return True

        if not self.location_match():
            return False

        # at this point we know that at least the name and location match
        if self.schema_match():
            return True

        # so the schemas are different.
        # The question remains if they are similar enough to allow reading.
        return self.schema_is_selectable()

    def schema_is_selectable(self):
        if self.base is None:
            return False
        # if differences are only in nuallability and comment,
        # we have no problem. This is handled implicitly.
        # If the specified schema is a subset of the storage schema,
        # then we can read without issue.

        # TODO: For now we do not support reading a
        #           subset of struct fields in the case of struct columns.
        #       Each top level column has to match in type.

        # Check if the specified schema is a subset of the disk schema.
        storage_schema_dict: Dict[str, StructField] = {
            field.name: field for field in self.base.schema.fields
        }
        for field in self.target.schema.fields:
            # every field has to be in the storage, or we can stop
            if field.name not in storage_schema_dict:
                return False

            # every type has to match
            if storage_schema_dict[field.name].dataType != field.dataType:
                return False

        # All specified fields are found with the correct dataType.
        # We can read the schema as-is (perhaps with a select statement
        # to ensure column capitalization and order)
        return True

    def _tblproperties_alter_statements(self) -> List[str]:
        """return that ALTER TABLE statements that will
        transform the tblproperties of the base table into the target"""
        statements = []
        if self.base.tblproperties == self.target.tblproperties:
            return statements

        updated_properties = {}
        for key, target_value in self.target.tblproperties.items():
            if key in self.base.tblproperties:
                if target_value == self.base.tblproperties[key]:
                    continue  # property is good
                else:
                    updated_properties[key] = target_value
            else:  # key not in base properties
                updated_properties[key] = target_value

        if updated_properties:
            parts = [
                f"{json.dumps(k)} = {json.dumps(v)}"
                for k, v in updated_properties.items()
            ]
            statements.append(
                f"ALTER TABLE {self.base.name} "
                f"""SET TBLPROPERTIES ({", ".join(parts)})"""
            )

        removed_properties = []
        for key in self.base.tblproperties.keys():
            if key not in self.target.tblproperties:
                removed_properties.append(key)
        if removed_properties:
            parts = [json.dumps(k) for k in removed_properties]
            statements.append(
                f"""ALTER TABLE {self.base.name} """
                f"""UNSET TBLPROPERTIES ({", ".join(parts)})"""
            )

        return statements

    def _schema_alter_statements(
        self,
        allow_columns_add=False,
        allow_columns_drop=False,
        allow_columns_type_change=False,
        errors_as_warnings=False,
        allow_columns_reorder=False,
    ) -> List[str]:
        if self.base.schema == self.target.schema:
            return []

        statements = []
        base_fields: Dict[str, StructField] = {
            field.name: field for field in self.base.schema.fields
        }
        base_order = [field.name for field in self.base.schema.fields]
        target_fields: Dict[str, StructField] = {
            field.name: field for field in self.target.schema.fields
        }
        target_order = [field.name for field in self.target.schema.fields]

        fields_to_add = []
        alter_comment = []  # (name,newValue)
        alter_nullability = []  # (name,newValue)
        fields_to_remove = []

        for key, targetField in target_fields.items():
            if key not in base_fields:
                if not allow_columns_add:
                    if errors_as_warnings:
                        print(f"WARNING: Would add column {key}. Continuing.")
                        continue
                    else:
                        raise TableSpectNotEnforcable(
                            f"use allow_columns_add to add {key}"
                        )

                fields_to_add.append(targetField)
                # adding the filed will fix the other field properties also
                continue

            # key is also in base
            base_field = base_fields[key]
            if base_field.dataType != targetField.dataType:
                if not allow_columns_type_change:
                    if errors_as_warnings:
                        print(
                            f"WARNING: Would change column {key}"
                            f" from {base_field.dataType}"
                            f" to {targetField.dataType}. Continuing."
                        )
                        continue
                    else:
                        raise TableSpectNotEnforcable(
                            f"set allow_columns_type_change to change {key}"
                        )

                fields_to_remove.append(key)
                fields_to_add.append(targetField)
                # adding the filed will fix the other field properties also
                continue

            # key is also in base and dataType matches
            base_comment = base_field.metadata.get("comment", "")
            target_comment = targetField.metadata.get("comment", "")
            if base_comment != target_comment:
                alter_comment.append((key, target_comment))

            if base_field.nullable != targetField.nullable:
                alter_nullability.append((key, targetField.nullable))

        for key in base_fields.keys():
            if key not in target_fields:
                if not allow_columns_drop:
                    if errors_as_warnings:
                        print(f"WARNING: Would drop column {key}. Continuing.")
                        continue
                    else:
                        raise TableSpectNotEnforcable(
                            f"use allow_columns_drop to drop {key}"
                        )
                fields_to_remove.append(key)

        # Now convert the accumulated changes to ALTER statements.
        if fields_to_remove:
            statement = f"ALTER TABLE {self.base.name} DROP COLUMN"
            if len(fields_to_remove) > 1:
                statement += "S"
            statement += " (" + ", ".join(fields_to_remove) + ")"

            statements.append(statement)

        if fields_to_add:
            statement = f"ALTER TABLE {self.base.name} ADD COLUMN"
            if len(fields_to_add) == 1:
                statement += (
                    " ("
                    + SchemaManager().struct_to_sql(StructType(fields_to_add)).strip()
                    + ")"
                )
            else:
                statement += "S"
                statement += (
                    " (\n  "
                    + SchemaManager().struct_to_sql(
                        StructType(fields_to_add), formatted=True
                    )
                    + "\n)"
                )

            statements.append(statement)

        for key, newValue in alter_nullability:
            statements.append(
                f"ALTER TABLE {self.base.name} ALTER COLUMN {key} "
                + ("SET" if newValue else "DROP")
                + " NOT NULL"
            )

        for key, newComment in alter_comment:
            statements.append(
                f"ALTER TABLE {self.base.name} ALTER COLUMN "
                f"{key} COMMENT {json.dumps(newComment)}"
            )

        # Finally address the possible change of order.
        base_order_so_far = copy.copy(base_order)
        for field in fields_to_add:
            base_order_so_far.append(field.name)

        # We may not have added and dropped all necessary columns, so we need to act
        # on the subset of columns that exist in both.
        desired_order = [col for col in target_order if col in base_order_so_far]
        base_order_so_far = [col for col in base_order_so_far if col in desired_order]

        # using set operations would not have preserved the order.
        alter_order_statements = []

        if base_order_so_far[0] != desired_order[0]:
            # need to update the first column
            base_order_so_far.pop(base_order_so_far.index(desired_order[0]))
            base_order_so_far.insert(0, desired_order[0])
            alter_order_statements.append(
                f"ALTER TABLE {self.base.name} ALTER COLUMN {desired_order[0]} FIRST"
            )

        for i, key in enumerate(desired_order[1:], 1):
            if base_order_so_far[i] != key:
                base_order_so_far.pop(base_order_so_far.index(key))
                base_order_so_far.insert(i, key)
                alter_order_statements.append(
                    f"ALTER TABLE {self.base.name} ALTER COLUMN {key} AFTER "
                    + desired_order[i - 1]
                )

        if alter_order_statements:
            if not allow_columns_reorder:
                if errors_as_warnings:
                    print("WARNING: Would reorder columns. Continuing.")
                else:
                    raise TableSpectNotEnforcable("use allow_columns_reorder if needed")
            else:
                statements += alter_order_statements

        return statements

    def _alter_statements_for_new_location(self, allow_columns_add=False) -> List[str]:
        """In case we need to change location,
        the alter statements will consist of creating the new target.
        and then pointing the current table at it. The target will be empty."""

        if not self.target.location:
            raise NotImplementedError(
                "Converting an external to a managed table not implemented"
            )

        # Note: This will fail if the target is not empty

        # Create the target table, then point the current table at it
        nameless_target = self.target.fully_substituted(name=None)
        statements = [nameless_target.get_sql_create()]
        statements.append(
            f"ALTER TABLE {self.base.name} "
            f"SET LOCATION {json.dumps(self.target.location)}"
        )
        # now to old base name points at the new location.
        # all that remains is to perhaps rename it
        new_base = nameless_target.fully_substituted(name=self.base.name)

        statements += DeltaTableSpecDifference(
            target=self.target, base=new_base
        ).alter_statements(allow_columns_add=allow_columns_add)
        return statements

    def alter_statements(
        self,
        allow_columns_add=False,
        allow_columns_drop=False,
        allow_columns_type_change=False,
        allow_columns_reorder=False,
        allow_name_change=False,
        allow_location_change=False,
        allow_table_create=True,
        errors_as_warnings=False,
    ) -> List[str]:
        """A list to alter statements that will ensure
        that what used to be the base table, becomes the target table"""
        target = self.target.fully_substituted()
        if self.base is None:
            if not (allow_table_create):
                if errors_as_warnings:
                    print(f"WARINING: Would create the table {target.name}")
                    return []
                else:
                    raise TableSpectNotEnforcable(
                        "use allow_table_create for creation " f"of {target.name}"
                    )
            else:
                return [target.get_sql_create()]

        full_diff = DeltaTableSpecDifference(
            target=target, base=self.base.fully_substituted()
        )

        base = full_diff.base
        target = full_diff.target

        if base.location != target.location:
            if allow_location_change:
                return full_diff._alter_statements_for_new_location(
                    allow_columns_add=allow_columns_add
                )
            else:
                if errors_as_warnings:
                    print(f"WARNING: Would change location of {base.name}. Continuing.")
                else:
                    raise TableSpectNotEnforcable(
                        "use allow_location_change if needed."
                    )

        statements = []

        # some table properties are needed to be able to drop columns.
        # therefore we need to set those first.
        statements += full_diff._tblproperties_alter_statements()

        statements += full_diff._schema_alter_statements(
            allow_columns_add=allow_columns_add,
            allow_columns_drop=allow_columns_drop,
            allow_columns_type_change=allow_columns_type_change,
            allow_columns_reorder=allow_columns_reorder,
            errors_as_warnings=errors_as_warnings,
        )

        if base.comment != target.comment:
            statements.append(
                f"""COMMENT ON TABLE {base.name} is {json.dumps(target.comment)}"""
            )

        # Change the name last so that we can still refer to the base name above this
        if base.name != target.name:
            if allow_name_change:
                statements.append(
                    f"""ALTER TABLE {base.name} RENAME TO {target.name}"""
                )
            else:
                if errors_as_warnings:
                    print(
                        f"WARNING: Would change name from {base.name} to {target.name}"
                    )
                else:
                    raise TableSpectNotEnforcable("set allow_name_change if needed.")

        return statements



import unittest

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.etl import Orchestrator
from spetlr.etl.extractors import StreamExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.loaders.stream_loader import StreamLoader
from spetlr.spark import Spark
from spetlr.sql import SqlHandle
from spetlr.testutils.stop_test_streams import stop_test_streams
from tests.cluster.sql.DeliverySqlServer import DeliverySqlServer


@unittest.skip("TODO: Test uses mount points")
class SqlServerStreamingTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        Configurator().clear_all_configurations()
        Configurator().set_debug()

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle.from_tc("MyDb").drop_cascade()
        DeliverySqlServer().drop_table("MSSQL")

        stop_test_streams()

    def test_01_configure(self):
        # Configure delta table
        tc = Configurator()
        tc.register(
            "MyDb", {"name": "TestDb{ID}", "path": "/mnt/spetlr/silver/testdb{ID}"}
        )

        tc.register(
            "MyTbl",
            {
                "name": "TestDb{ID}.TestTbl",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtbl",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}/_checkpoint_path_tbl",
                "query_name": "testquerytbl{ID}",
            },
        )

        DbHandle.from_tc("MyDb").create()

        dh = DeltaHandle.from_tc("MyTbl")

        Spark.get().sql(
            f"""
               CREATE TABLE {dh.get_tablename()}
               (
               testcolumn int
               )
               LOCATION '{Configurator().get("MyTbl", "path")}'
           """
        )

        df = Spark.get().createDataFrame([(1,), (2,)], "testcolumn int")

        dh.overwrite(df, mergeSchema=True)

        # Configure sql table
        tc.register(
            "MSSQL",
            {
                "name": "dbo.stream_test{ID}",
                "checkpoint_path": "/mnt/spetlr/silver/stream_test_sql{ID}"
                "/_checkpoint_path_",
                "query_name": "testquerysql{ID}",
            },
        )

        _table = Configurator().table_name("MSSQL")

        sql_argument = f"""
                    IF OBJECT_ID('{_table}', 'U') IS NULL
                    BEGIN
                    CREATE TABLE {_table}
                    (
                        testcolumn INT NULL
                    )
                    END
                """

        DeliverySqlServer().execute_sql(sql_argument)

    def test_02_stream_to_sql(self):
        _table = Configurator().table_name("MSSQL")
        sql_handle = SqlHandle(_table, sql_server=DeliverySqlServer())
        dh = DeltaHandle.from_tc("MyTbl")

        o = Orchestrator()
        o.extract_from(StreamExtractor(dh, dataset_key="MyTbl"))
        o.load_into(
            StreamLoader(
                loader=SimpleLoader(sql_handle, mode="append"),
                await_termination=True,
                checkpoint_path=Configurator().get("MSSQL", "checkpoint_path"),
                query_name=Configurator().get("MSSQL", "query_name"),
            )
        )
        o.execute()

        result = sql_handle.read()
        self.assertEqual(2, result.count())



import datetime
from datetime import datetime as dt
from typing import List, Tuple

import py4j.protocol
import pyspark.sql.utils
from pyspark.sql import DataFrame
from pyspark.sql import functions as f

from spetlr.configurator.configurator import Configurator
from spetlr.spark import Spark

utc = datetime.timezone.utc


class EventHubCaptureExtractor:
    """Class to access eventhub capture files as table.

    The constructor argument should be given as a key that can be found in
    the Table Configurator, where the relevant table must be marked with
    format=avro and with a partitioning from the set of known partitions: y,m,d,h
    """

    path: str
    partitioning: str

    @classmethod
    def from_tc(cls, tbl_id: str):
        tc = Configurator()
        assert tc.get(tbl_id, "format") == "avro"
        return cls(
            path=tc.get(tbl_id, "path"),
            partitioning=tc.get(tbl_id, "partitioning"),
        )

    def __init__(self, path: str, partitioning: str):
        self.path = path
        self.partitioning = partitioning.lower()
        assert self.partitioning in ["ymd", "ymdh"]

    def _validate_timestamp(self, stamp: dt):
        """Check that the given timestamp is an edge
        between two capture periods according to the set partitioning."""
        stamp = stamp.astimezone(utc)
        if stamp.microsecond != 0:
            return False
        if stamp.second != 0:
            return False
        if stamp.minute != 0:
            return False
        if "h" not in self.partitioning and stamp.hour != 0:
            return False
        return True

    def _add_columns(self, df: DataFrame) -> DataFrame:
        # here we extract the partition description from the input filename
        # partitions are saved in folders like .../y=2022/m=09/d=23/h=02/...
        # there are many ways to extract these parts. We use a single regular expression
        # the parts are ordered by assumption (validated in manual tests)
        # but this ordering is not verified in order to prioritize performance
        df = df.withColumn(
            "_parts",
            f.expr('regexp_extract_all(input_file_name(),"[ymdh]=([0-9]+)/",1)'),
        )
        df = df.withColumn("y", f.element_at("_parts", 1).cast("INTEGER"))
        df = df.withColumn("m", f.element_at("_parts", 2).cast("INTEGER"))
        df = df.withColumn("d", f.element_at("_parts", 3).cast("INTEGER"))
        if "h" in self.partitioning:
            df = df.withColumn("h", f.element_at("_parts", 4).cast("INTEGER"))

        # unfortunately, regexp_extract_all and make_timestamp are not exposed
        # in the pyspark wrapper library
        df = df.withColumn(
            "pdate",
            f.expr(
                "make_timestamp(y,m,d,"
                + ("h," if "h" in self.partitioning else "0,")
                + '0,0,"UTC")'
            ),
        )
        df = df.drop("_parts")

        # for some bizarre reason, the built-in string timestamp uses
        # a localized date-time format and not a standardized one.
        # add a standardized timestamp
        df = df.withColumn(
            "EnqueuedTimestamp",
            f.to_timestamp(f.col("EnqueuedTimeUtc"), "M/d/yyyy h:mm:ss a"),
        )
        return df

    def _now_utc(self):
        # this method needs to be here so that I can override it in the unittest
        return dt.now(tz=utc)

    def _validated_slice_arguments(
        self, from_partition: dt = None, to_partition: dt = None
    ) -> Tuple[dt, dt]:
        from_partition = from_partition.astimezone(datetime.timezone.utc)

        if not self._validate_timestamp(from_partition):
            raise ValueError("from_partition is not a pure partition edge")

        if to_partition is None:
            # if the to_partition is not given, we set it to the end of
            to_partition = dt(self._now_utc().year + 1, 1, 1, tzinfo=utc)
        else:
            # to_partition was given by the caller, we need to validate it.
            to_partition = to_partition.astimezone(datetime.timezone.utc)

            if not self._validate_timestamp(to_partition):
                raise ValueError("to_partition is not a clean partition edge")

            # ensure that limits are ordered correctly
            if to_partition < from_partition:
                raise ValueError("Negative partition interval.")

        if to_partition == from_partition:
            raise ValueError("Requested to read over zero partitions.")

        return from_partition, to_partition

    def _break_into_partitioning_parts(self, fp: dt, tp: dt) -> List[str]:
        """
        data is grouped to use as few spark.read.load commands as possible
        while still respecting the given limits exactly.
        The slice is cut using the largest bite possible for each cut.
        """
        parts_to_load: List[str] = []

        # We now have a slice description [fp, tp)
        # we now take bites out of this slice until the slice is closed.
        # in each iteration we take the biggest bite that we can,
        while fp != tp and fp < self._now_utc():
            # consider a year-sized bite
            # does fp represent a lower edge of a year slice:
            if fp.month == 1 and fp.day == 1 and fp.hour == 0:
                # fp seems suitable, check tp
                year_end = fp.replace(year=fp.year + 1)
                if year_end <= tp:
                    # tp is far enough away. We can take a one-year bite!
                    parts_to_load.append(f"y={fp.year:04}/")
                    fp = year_end
                    continue

            # consider a month-sized bite
            if fp.day == 1 and fp.hour == 0:
                # fp seems suitable, check tp
                # no easier way to add a month afaik
                month_end = (fp + datetime.timedelta(days=32)).replace(day=1)
                if month_end <= tp:
                    # we can take a one-month bite!
                    parts_to_load.append(f"y={fp.year:04}/m={fp.month:02}/")
                    fp = month_end
                    continue

            # consider a day-sized bite
            if fp.hour == 0:
                # fp seems suitable, check tp
                day_end = fp + datetime.timedelta(days=1)
                if day_end <= tp:
                    # we can take a one-month bite!
                    parts_to_load.append(
                        f"y={fp.year:04}/m={fp.month:02}/d={fp.day:02}/"
                    )
                    fp = day_end
                    continue

            # consider an hour-sized bite
            if "h" in self.partitioning:
                # fp is always suitable due to argument validation
                hour_end = fp + datetime.timedelta(hours=1)
                # we don't need really need to check tp
                # because we know that fp!=tp and that fp<tp and that
                # both represent valid partitioning edges
                # but let's do it anyway
                if hour_end <= tp:
                    # we can take a one-hour bite!
                    parts_to_load.append(
                        f"y={fp.year:04}/m={fp.month:02}/"
                        f"d={fp.day:02}/h={fp.hour:02}/"
                    )
                    fp = hour_end
                    continue

            raise Exception("The loading logic failed. Contact the maintainer.")
        return parts_to_load

    def _load_union_of_parts(self, parts_to_load: List[str]) -> DataFrame:
        spark = Spark.get()
        df = None
        for part in parts_to_load:
            try:
                part_df = spark.read.format("avro").load(self.path + "/" + part)
            except (pyspark.sql.utils.AnalysisException, py4j.protocol.Py4JError):
                print(
                    f"WARNING: part {part} caused an analysis exception. "
                    "The partition is probably empty."
                )
                continue
            if df is None:
                df = self._add_columns(part_df)
            else:
                df = df.unionByName(self._add_columns(part_df))

        # Note: return value can be None!
        return df

    def read(
        self,
        from_partition: datetime.datetime = None,
        to_partition: datetime.datetime = None,
    ) -> DataFrame:
        r"""Read all data between the two limits.
        Supported usage:
            - set neither value, all files are read.
            - set from_partition only, (to_partition is assumed to be now)
            - set both from_partition and to_partition
        Given partition limits are designating the edges of partitions to read. Hence:
            - The from_partition is inclusive.
            - The to_partition is exclusive.
        If no to_partition is given, all data will be read, including the
        (usually only partially filled) current partition.
        """

        # internally there are only two real cases,
        # reading everything
        # or reading a slice.
        # first we handle the case of reading everything
        if from_partition is None:
            if to_partition is not None:
                raise ValueError("Use both limits or only 'from' limit.")
            else:
                # read all
                return self._add_columns(
                    Spark.get().read.format("avro").load(self.path)
                )

        # now we know that a slice will be read.

        # we need pure inputs. The timestamps have to point at partition edges
        from_partition, to_partition = self._validated_slice_arguments(
            from_partition, to_partition
        )

        # data is grouped to use as few spark.read.load commands as possible
        # while still respecting the given limits exactly
        parts_to_load = self._break_into_partitioning_parts(
            from_partition, to_partition
        )

        # we know that from_partition != to_partition
        # and the function above must return some partitions
        assert len(parts_to_load)

        df = self._load_union_of_parts(parts_to_load)
        if df is None:
            schema = Spark.get().read.format("avro").load(self.path).schema
            df = Spark.get().createDataFrame([], schema)

        return df

    def get_partitioning(self):
        return list(self.partitioning)



import importlib.resources
import uuid
from pathlib import Path
from string import Formatter
from types import ModuleType
from typing import Any, Dict, Set, Union

import yaml
from deprecated import deprecated

from spetlr.configurator._cli.ConfiguratorCli import ConfiguratorCli
from spetlr.configurator.sql import _parse_sql_to_config
from spetlr.exceptions import NoSuchValueException
from spetlr.functions import json_hash

# recursive type definition of the details object
TcDetails = Dict[str, Union[str, "TcDetails"]]
TcValue = Union[str, TcDetails]


class ConfiguratorSingleton(type):
    """The reason that we do not use spetlr.singleton here,
    is that the behavior of that metaclass depends on the classname.
    Which prevents us from making a deprecated subclass of the old name
    which actually instantiates to the new class instance."""

    _instance = None

    def __call__(cls, *args, **kwargs):
        if ConfiguratorSingleton._instance is None:
            ConfiguratorSingleton._instance = super(
                ConfiguratorSingleton, cls
            ).__call__(*args, **kwargs)
        return ConfiguratorSingleton._instance


class Configurator(ConfiguratorCli, metaclass=ConfiguratorSingleton):
    _DEFAULT = object()
    _unique_id: str
    _raw_resource_details: TcDetails
    _is_debug: bool

    # this dict contains all details for all resources
    table_details: Dict[str, str]

    def __init__(
        self,
        resource_path: Union[str, ModuleType] = None,
    ):
        self._unique_id = uuid.uuid4().hex
        self.clear_all_configurations()

        if resource_path:
            self.add_resource_path(resource_path)

    def all_keys(self):
        """All keys that appear in the configuration files."""
        return list(self._raw_resource_details.keys())

    def clear_all_configurations(self):
        self._raw_resource_details = dict()
        self._is_debug = False
        self.table_details = dict()
        self._set_extras()

    def verify_consistency(self):
        """This method will re-build the table details. This requires that all
        internally recursive references can be resolved."""
        self.table_details = dict()
        return self.get_all_details()

    ############################################
    # the core logic of this class is contained
    # in the following methods
    ############################################

    def _set_extras(self):
        self.register("ID", {"release": "", "debug": f"__{self._unique_id}"})
        self.register("MNT", {"release": "mnt", "debug": "tmp"})

    @deprecated(
        reason="use .get('ENV') to get literal values.",
    )
    def get_extra_details(self) -> Dict[str, str]:
        """The distinction between extras and normal values was removed.
        This method will always return an empty dict. Extra settings are now
        string items. Retrieve them with .get(id)"""
        return {}

    def _get_item(self, table_id: str) -> TcValue:
        """item dictionary where release-debug and alias loops are resolved"""
        # explanation to the maintainer:
        # There are many ways to specify a value in the Configurator:
        # MyLiteral: "MyValue"
        #
        # MyKey:
        #   name: someName
        #   path: some/Path
        #
        # MyAlias:
        #   alias: MyKey
        #
        # MyForked:
        #   release:
        #     alias: MyAlias
        #   debug:
        #     name: anotherName
        #     path: another/Path
        #
        # What we want is to support that in all cases we get the intended value out
        # tc._get_item("MyLiteral") -> "MyValue"
        # tc._get_item("MyKey") -> {'name':..., 'path':...}
        # tc._get_item("MyAlias") -> same as tc._get_item("MyKey")
        # tc._get_item("MyForked")
        #   if not self._is_debug -> same as tc._get_item("MyKey")
        #   if self._is_debug -> {'name':'anotherName', 'path':'another/Path'}
        #
        # resolving forks and aliases is the task of this method.

        # this stack allows us to detect alias loops
        stack = {table_id}

        value: TcValue = self._raw_resource_details[table_id]

        while True:
            if not isinstance(value, dict):
                # situation like MyLiteral
                return value
            else:
                # value is a dict
                if set(value.keys()) == {"release", "debug"}:
                    # Situation like MyForked
                    if self._is_debug:
                        value = value["debug"]
                    else:
                        value = value["release"]
                    continue
                elif set(value.keys()) == {"alias"}:  # allow alias of alias
                    # Situation like MyAlias
                    new_id = value["alias"]
                    if new_id in stack:
                        raise ValueError(f"Alias loop at key {new_id}")
                    stack.add(new_id)
                    value = self._raw_resource_details[new_id]
                    continue
                else:
                    # value is a dict,
                    # but it is neither a forking (release,debug)
                    # nor is it an alias
                    # hence we have arrived at the final value set of this item
                    # This is how both MyForked and MyAlias can arrive at MyKey
                    return value

    def _get_unsubstituted_item_property(self, table_id: str, property: str) -> str:
        """item property where release-debug and alias loops are resolved"""
        value = self._get_item(table_id)
        if not property:
            if isinstance(value, dict):
                raise NoSuchValueException("Cannot get bare string. Item is a Dict")
            else:
                return value
        if property in value:
            return value[property]
        else:
            raise NoSuchValueException(property)

    def _get_item_property(
        self, table_id: str, property: str, _forbidden_keys: Set[str] = None
    ) -> str:
        """Get the full item property, fully resolved and substituted."""
        raw_string = self._get_unsubstituted_item_property(table_id, property)

        # some items are not strings, then the rest of this function makes no sense
        if not isinstance(raw_string, str):
            return raw_string

        # get all keys used in the raw_string, such as using {MyDb} will get "MyDb"
        format_keys = set(
            i[1] for i in Formatter().parse(raw_string) if i[1] is not None
        )

        # the forbidden-keys logic allows us to detect reference loops.
        # no key that is in the upstream of a property is allowed in the string
        # substitutions of this property.
        _forbidden_keys = _forbidden_keys or set()
        composite_key = table_id
        if property:
            composite_key += f"_{property}"
        _forbidden_keys.add(composite_key)
        if property == "name":
            _forbidden_keys.add(table_id)
        if any(key in _forbidden_keys for key in format_keys):
            raise ValueError(
                f"Substitution loop at table {table_id} property {property}"
            )

        # every key that is not in the special properties and not forbidden
        # should refer to another resource. If it does not an exception will occur.
        replacements = {}
        for key in format_keys:
            try:
                # maybe we have a case of reference to property, let's try:

                id_part, property_part = key.rsplit("_", 1)
                # will raise ValueError if there are too few parts,
                # keys without _ are handled below

                value = self._get_item_property(
                    id_part, property_part, _forbidden_keys.copy()
                )
                # raises ValueError if it does not exist

                replacements[key] = value
                continue
            except ValueError:
                pass

            # if we get here, there was no _ in the key. Either the key exists as a bare
            # string value, try that:
            try:
                replacements[key] = self._get_item_property(key, "", _forbidden_keys)
                continue
            except NoSuchValueException:
                pass

            # otherwise bare key references are to 'name',
            # which _must_ exist in this case
            replacements[key] = self._get_item_property(key, "name", _forbidden_keys)

        # we have run through the key names of all replacement keys in the string.
        # Any that we could not find were skipped silently above, but that means that
        # we do not have them in 'replacements'. Therefore, this next step will raise an
        # exception for any missing key, which will give a meaningful error to the user.
        return raw_string.format(**replacements)

    def add_resource_path(
        self, resource_path: Union[str, ModuleType], consistency_check=True
    ) -> None:
        self.table_details = dict()
        backup_details = self._raw_resource_details.copy()
        try:
            for file_name in importlib.resources.contents(resource_path):
                extension = Path(file_name).suffix
                if extension not in [".json", ".yaml", ".yml"]:
                    continue
                with importlib.resources.path(resource_path, file_name) as file_path:
                    with open(file_path) as file:
                        # just use yaml since json is a subset of yaml
                        update = yaml.load(file, Loader=yaml.FullLoader)

                        if not isinstance(update, dict):
                            raise ValueError(f"document in {file_path} is no dict.")

                        for key, value in update.items():
                            # we now support all bare value types in yaml.
                            # no further checking
                            self.register(key, value)

            # try re-building all details
            if consistency_check:
                self.verify_consistency()
        except:  # noqa: E722  we re-raise the exception, so bare except is ok.
            # this piece makes it so that the Configurator can still be used
            # if any exception raised by the above code is caught.
            self._raw_resource_details = backup_details
            raise

    def add_sql_resource_path(
        self, resource_path: Union[str, ModuleType], consistency_check=True
    ) -> None:
        self.table_details = dict()
        for key, value in _parse_sql_to_config(resource_path).items():
            self.register(key, value)

        if consistency_check:
            self.verify_consistency()

    def key_of(self, attribute: str, value: str) -> str:
        """Obtain the key of the first registered item that has a given attribute
        set to a given value. Uniqueness of the match is the responsibility of
        the library user.

        This function is slow as it performs a linear search. It is intended for use in
        setup code.
        """
        for key, object in self._raw_resource_details.items():
            try:
                if object[attribute] == value:
                    return key
            except (KeyError, TypeError):
                continue
        else:
            raise KeyError(f"No key with attribute {attribute}={repr(value)}")

    ############################################
    # all methods below are interface and convenience methods
    ############################################

    @deprecated(
        reason="register literal string values instead.",
    )
    def set_extra(self, **kwargs: str):
        """Use .register(key,value) instead.
        for example call .register('ENV','prod')"""
        for key, value in kwargs.items():
            self.register(key, value)

    def set_debug(self):
        """Select debug tables. {ID} will be replaced with a guid"""
        self.reset(debug=True)

    def set_prod(self):
        """Select production tables. {ID} will be replaced with a empty string"""
        self.reset(debug=False)

    def __reset(self, debug: bool) -> None:
        self._is_debug = debug
        self.table_details = dict()

    def reset(self, *, debug: bool = False):
        """
        Resets table names and table SQL. Enables or disables debug mode
        (used for unit tests and integration tests).
        :param debug: False -> release tables, True -> debug tables.
        :param kwargs: additional keys to be substituted in names and paths
        """
        self.__reset(debug)

    def is_debug(self):
        """
        Return True if table names and table SQL specify debug table,
        False if release tables
        """
        return self._is_debug

    def get_unique_id_length(self):
        """
        Return the character length of the UUID identifier inserted into
        names with the {ID} tag
        """
        return len(self._unique_id)

    def register(self, key: str, value: TcValue) -> str:
        """
        Register a new item and return its key.
        If both the new and old items are dictionaries, their contents are merged.
        Supply value=None to clear a key.
        """
        if value is None:
            self._raw_resource_details.pop(key, None)
        elif isinstance(value, dict) and isinstance(
            self._raw_resource_details.get(key), dict
        ):
            # if both are dicts, update the contents.
            self._raw_resource_details[key].update(value)

            # This is how to clear a key if the keyvalue is None
            self._raw_resource_details[key] = {
                k: v
                for k, v in self._raw_resource_details[key].items()
                if v is not None
            }

            # If the key has no remaining values, remove it completely
            if not self._raw_resource_details[key]:
                self._raw_resource_details.pop(key)

        else:
            self._raw_resource_details[key] = value
        self.table_details = dict()
        return key

    def define(self, **kwargs) -> str:
        """
        Register a new item based on its properties only and return its key.
        The returned key is a hash-like string depending only on the values.
        """
        if not kwargs:
            raise ValueError("No value passed.")

        key = json_hash(kwargs)
        return self.register(key, kwargs)

    @deprecated(
        reason="Use .get(table_id,property_name) instead.",
    )
    def table_property(
        self, table_id: str, property_name: str, default_value: str = None
    ):
        """
        Return the table property (e.g. name, path, format, etc.)
            for the specified table id.
        :param table_id: Table id in the .json or .yaml files.
        :param property_name: Name of the property to read (e.g. "name", "path", etc.)
        :param default_value: Optional default value of the property
            if the property is missing.
        :return: str: property value
        """
        property_value = self.get_all_details().get(
            f"{table_id}_{property_name}", default_value
        )

        if property_value is None:
            raise ValueError(
                f"property '{property_name}' for table identifier '{table_id}' is empty"
            )
        return property_value

    def table_name(self, table_id: str):
        """
        Return the table name for the specified table id.
        :param table_id: Table id in the .json or .yaml files.
        :return: str: table name
        """
        return self.get(table_id, "name")

    @deprecated(
        reason='Use .get(table_id,"path") instead.',
    )
    def table_path(self, table_id: str):
        """
        Return the table path for the specified table id.
        :param table_id: Table id in the .json or .yaml files.
        :return: str: table path
        """
        return self.get(table_id, "path")

    def get(self, table_id: str, property: str = "", default: Any = _DEFAULT):
        """return the property of the table_id.
        To get raw strings, specify no property.
        If default is set, it is returned instead of raising in case of missing keys.
        Return value will be whatever was registered under the given property."""
        try:
            return self._get_item_property(table_id, property)
        except NoSuchValueException:
            if default is self._DEFAULT:
                raise
            else:
                return default

    def get_all_details(self):
        """
        Return a dictionary containing every resource detail fully resolved.
        e.g. a resource that looks like this:
        MyTableDetail:
            name: mytablename
            path: my/table/path
        will appear with three keys:
            - "MyTableDetail"
            - "MyTableDetail_name"
            - "MyTableDetail_path"
        all substitutions will be fully resolved.
        """
        if not self.table_details:
            self.table_details = dict()

            for table_id in self._raw_resource_details.keys():
                # add the name as the bare key
                try:
                    self.table_details[table_id] = self.get(table_id)
                    continue  # if it was a bare string, we can stop here
                except NoSuchValueException:
                    pass

                try:
                    self.table_details[table_id] = self.get(table_id, "name")
                except NoSuchValueException:
                    pass

                # add every property as a _property part
                for property_name in set(self._get_item(table_id).keys()):
                    try:
                        item = self.get(table_id, property_name)
                    except NoSuchValueException:
                        continue
                    # if the dict values are dicts, stop here,
                    # not supported for direct substitution
                    # this will take care of definitions of schema and similar.
                    if not isinstance(item, dict):
                        self.table_details[f"{table_id}_{property_name}"] = str(item)

        return self.table_details

    def regenerate_unique_id_and_clear_conf(self):
        self._unique_id = uuid.uuid4().hex
        self.clear_all_configurations()
